import os
from typing import Any, Optional

import google.generativeai as genai
from google.generativeai.types.content_types import ContentType
from google.generativeai.types.generation_types import GenerationConfig


def get_genai_model(
    model_name: str,
    api_key_env_name: str,
    n_results: int = 1,
    system_instruction: Optional[ContentType] = None,
    max_output_tokens: int = 2048,
    json_mode: bool = False,
    response_schema: dict[str, Any] | None = None,
    temperature: float | None = None,
    top_p: float | None = None,
    top_k: int | None = None,
):
    if api_key_env_name is None:
        raise ValueError("api_key_env_name must be provided to use the Gemini API.")

    generation_config = GenerationConfig(
        candidate_count=n_results,
        stop_sequences=None,
        max_output_tokens=max_output_tokens,
        temperature=temperature,
        top_p=top_p,
        top_k=top_k,
        response_mime_type="text/plain" if not json_mode else "application/json",
        response_schema=response_schema,
    )

    genai.configure(api_key=os.getenv(api_key_env_name))
    return genai.GenerativeModel(
        model_name=model_name,
        system_instruction=system_instruction,
        generation_config=generation_config,
    )



from typing import Any, Callable, Iterable, Literal, Mapping, Union

import httpx
from anthropic._base_client import DEFAULT_MAX_RETRIES
from anthropic._types import (
    NOT_GIVEN,
    NotGiven,
    ProxiesTypes,
    Timeout,
    Transport,
)
from anthropic.types.text_block_param import TextBlockParam

from ..base import (
    BaseConversationLengthAdjuster,
    BaseConversationMemory,
    BaseFilter,
)
from ..base_assembly import BaseAssembly
from ..message_content import ConversationType, InputType, Message, TextContent, ToolContent
from ..result import CompletionResults, FunctionCallingResults
from ..tools import ToolConfig
from ..usage import TokenCounter, Usage
from .llm.chat import AnthropicChatModule
from .llm.function_calling import AnthropicFunctionCallingModule


class ClaudeFunctionalChat(BaseAssembly):
    """
    FIXME : This module might not work for multi-turn conversation using tools.
    """

    def __init__(
        self,
        model_name: str,
        api_type: str = "anthropic",
        api_key_env_name: str | None = None,
        auth_token_env_name: str | None = None,
        endpoint_env_name: str | httpx.URL | None = None,
        aws_secret_key_env_name: str | None = None,
        aws_access_key_env_name: str | None = None,
        aws_region_env_name: str | None = None,
        aws_session_token_env_name: str | None = None,
        gc_region_env_name: str | NotGiven = NOT_GIVEN,
        gc_project_id_env_name: str | NotGiven = NOT_GIVEN,
        gc_access_token_env_name: str | None = None,
        credentials: Any | None = None,
        timeout: Union[float, Timeout, None, NotGiven] = 600,
        max_retries: int = DEFAULT_MAX_RETRIES,
        default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        http_client: httpx.Client | None = None,
        transport: Transport | None = None,
        proxies: ProxiesTypes | None = None,
        connection_pool_limits: httpx.Limits | None = None,
        _strict_response_validation: bool = False,
        conversation_length_adjuster: BaseConversationLengthAdjuster | None = None,
        system_instruction: Union[str, Iterable[TextBlockParam]] | NotGiven = NOT_GIVEN,
        conversation_memory: BaseConversationMemory | None = None,
        content_filter: BaseFilter | None = None,
        token_counter: TokenCounter | None = None,
        tools: list[Callable] | None = None,
        tool_configs: list[ToolConfig] | None = None,
        temperature: float | NotGiven = NOT_GIVEN,
        top_k: int | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven = NOT_GIVEN,
    ):
        super().__init__(conversation_memory=conversation_memory)

        self.chat = AnthropicChatModule(
            model_name=model_name,
            api_type=api_type,
            api_key_env_name=api_key_env_name,
            auth_token_env_name=auth_token_env_name,
            endpoint_env_name=endpoint_env_name,
            aws_secret_key_env_name=aws_secret_key_env_name,
            aws_access_key_env_name=aws_access_key_env_name,
            aws_region_env_name=aws_region_env_name,
            aws_session_token_env_name=aws_session_token_env_name,
            gc_region_env_name=gc_region_env_name,
            gc_project_id_env_name=gc_project_id_env_name,
            gc_access_token_env_name=gc_access_token_env_name,
            credentials=credentials,
            timeout=timeout,
            max_retries=max_retries,
            default_headers=default_headers,
            default_query=default_query,
            http_client=http_client,
            transport=transport,
            proxies=proxies,
            connection_pool_limits=connection_pool_limits,
            _strict_response_validation=_strict_response_validation,
            conversation_length_adjuster=conversation_length_adjuster,
            system_instruction=system_instruction,
            conversation_memory=self.conversation_memory,
            content_filter=content_filter,
            token_counter=token_counter,
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
        )

        if tools:
            self.function_calling = AnthropicFunctionCallingModule(
                model_name=model_name,
                api_type=api_type,
                api_key_env_name=api_key_env_name,
                auth_token_env_name=auth_token_env_name,
                endpoint_env_name=endpoint_env_name,
                aws_secret_key_env_name=aws_secret_key_env_name,
                aws_access_key_env_name=aws_access_key_env_name,
                aws_region_env_name=aws_region_env_name,
                aws_session_token_env_name=aws_session_token_env_name,
                gc_region_env_name=gc_region_env_name,
                gc_project_id_env_name=gc_project_id_env_name,
                gc_access_token_env_name=gc_access_token_env_name,
                credentials=credentials,
                timeout=timeout,
                max_retries=max_retries,
                default_headers=default_headers,
                default_query=default_query,
                http_client=http_client,
                transport=transport,
                proxies=proxies,
                connection_pool_limits=connection_pool_limits,
                _strict_response_validation=_strict_response_validation,
                conversation_length_adjuster=conversation_length_adjuster,
                system_instruction=system_instruction,
                conversation_memory=self.conversation_memory,
                content_filter=content_filter,
                token_counter=token_counter,
                tools=tools,
                tool_configs=tool_configs,
                temperature=temperature,
                top_k=top_k,
                top_p=top_p,
            )
        else:
            self.function_calling = None

    def run(
        self,
        prompt: InputType,
        init_conversation: ConversationType | None = None,
        tool_choice: Literal["auto", "any"] | str | None = None,
        tool_only: bool = False,
    ) -> CompletionResults | FunctionCallingResults:
        if self.function_calling and tool_choice is None:
            tool_choice = "auto"

        if tool_only:
            assert tool_choice is not None, "tool_choice must be provided when tool_only is True"

        if self.function_calling:
            if tool_choice is not None:
                total_usage = Usage(model_name=self.chat.chat_model.model_name)

                response_function_calling: FunctionCallingResults = self.function_calling.run(
                    prompt,
                    init_conversation=init_conversation,
                    tool_choice=tool_choice,
                )

                if tool_only:
                    self._clear_memory()

                    return response_function_calling

                total_usage += response_function_calling.usage

                if response_function_calling.results:
                    prompt = Message(
                        role="function",
                        content=[
                            ToolContent(**content.model_dump())
                            for result in response_function_calling.results
                            for content in result.content
                        ],
                    )
                    init_conversation = (
                        None  # if tool is used, init_conversation is stored in the memory
                    )

                    response_function_calling: FunctionCallingResults = self.function_calling.run(
                        prompt, init_conversation=init_conversation
                    )

                    total_usage += response_function_calling.usage

                self._clear_memory()

                return CompletionResults(
                    usage=total_usage,
                    prompt=response_function_calling.prompt,
                    message=Message(
                        role="assistant",
                        content=[
                            c
                            for c in response_function_calling.calls.content
                            if isinstance(c, TextContent)
                        ],
                    ),
                )

            else:
                raise ValueError("tool_choice must be provided when the model has tools available")
        else:
            response_chat: CompletionResults = self.chat.run(
                prompt, init_conversation=init_conversation
            )

            self._clear_memory()

            return response_chat

    async def arun(
        self,
        prompt: InputType,
        init_conversation: ConversationType | None = None,
        tool_choice: Literal["auto", "any"] | str | None = None,
        tool_only: bool = False,
    ) -> CompletionResults | FunctionCallingResults:
        if self.function_calling and tool_choice is None:
            tool_choice = "auto"

        if tool_only:
            assert tool_choice is not None, "tool_choice must be provided when tool_only is True"

        if self.function_calling:
            if tool_choice is not None:
                total_usage = Usage(model_name=self.chat.chat_model.model_name)

                response_function_calling: FunctionCallingResults = (
                    await self.function_calling.arun(
                        prompt,
                        init_conversation=init_conversation,
                        tool_choice=tool_choice,
                    )
                )

                if tool_only:
                    self._clear_memory()

                    return response_function_calling

                total_usage += response_function_calling.usage

                if response_function_calling.results:
                    prompt = Message(
                        role="function",
                        content=[
                            ToolContent(**content.model_dump())
                            for result in response_function_calling.results
                            for content in result.content
                        ],
                    )
                    init_conversation = (
                        None  # if tool is used, init_conversation is stored in the memory
                    )

                    response_function_calling: FunctionCallingResults = (
                        await self.function_calling.arun(
                            prompt, init_conversation=init_conversation
                        )
                    )

                    total_usage += response_function_calling.usage

                self._clear_memory()

                return CompletionResults(
                    usage=total_usage,
                    prompt=response_function_calling.prompt,
                    message=Message(
                        role="assistant",
                        content=[
                            c
                            for c in response_function_calling.calls.content
                            if isinstance(c, TextContent)
                        ],
                    ),
                )
            else:
                raise ValueError("tool_choice must be provided when the model has tools available")
        else:
            response_chat: CompletionResults = await self.chat.arun(
                prompt, init_conversation=init_conversation
            )

            self._clear_memory()

            return response_chat



import json
from typing import Any, overload

from anthropic.types import (
    ImageBlockParam,
    MessageParam,
    TextBlockParam,
    ToolResultBlockParam,
    ToolUseBlockParam,
)
from anthropic.types.image_block_param import Source

from ..base import BaseMessage, FunctionCallingResults, ToolCallResponse
from ..message_content import ImageContent, Message, TextContent, ToolCall, ToolContent
from ..utils import decode_image


class ClaudeMessage(BaseMessage):
    @property
    def as_user(self) -> MessageParam:
        return MessageParam(role="user", content=self.contents)

    @property
    def as_assistant(self) -> MessageParam:
        return MessageParam(role="assistant", content=self.contents)

    @property
    def as_function(self) -> MessageParam:
        return MessageParam(role="user", content=self.contents)

    @property
    def as_function_call(self) -> MessageParam:
        return MessageParam(role="assistant", content=self.contents)

    @staticmethod
    def _format_text_content(content: TextContent) -> TextBlockParam:
        return TextBlockParam(text=content.text, type="text")

    @staticmethod
    def _format_image_content(content: ImageContent) -> ImageBlockParam:
        file_format = decode_image(content.image, as_utf8=True).format.lower()

        return ImageBlockParam(
            type="image",
            source=Source(data=content.image, media_type=f"image/{file_format}", type="base64"),
        )

    @staticmethod
    def _format_tool_content(content: ToolContent) -> ToolResultBlockParam:
        return ToolResultBlockParam(
            tool_use_id="toolu_" + content.call_id.split("_")[-1],
            type="tool_result",
            content=content.output,
        )

    @overload
    @staticmethod
    def _format_tool_call_content(content: TextContent) -> TextBlockParam: ...

    @overload
    @staticmethod
    def _format_tool_call_content(content: ToolCall) -> ToolUseBlockParam: ...

    @staticmethod
    def _format_tool_call_content(
        content: TextContent | ToolCall,
    ) -> TextBlockParam | ToolUseBlockParam:
        if isinstance(content, TextContent):
            return TextBlockParam(text=content.text, type="text")
        elif isinstance(content, ToolCall):
            return ToolUseBlockParam(
                id="toolu_" + content.call_id.split("_")[-1],
                type="tool_use",
                input=content.args if isinstance(content.args, dict) else json.loads(content.args),
                name=content.name,
            )
        else:
            raise ValueError("Invalid content type")

    @classmethod
    def from_client_message(cls, message: Any) -> Message:
        serializable = cls._to_dict(message)

        return Message(
            role=serializable.get("role"),
            content=serializable.get("content"),
            name=serializable.get("name"),
        )

    @staticmethod
    def _to_dict(message: MessageParam | dict) -> dict[str, Any]:
        return {
            "role": message["role"],
            "content": [c.to_dict() if hasattr(c, "to_dict") else c for c in message["content"]],
        }

    @classmethod
    def to_universal_message_from_function_call(cls, response: FunctionCallingResults) -> Message:
        is_only_text = all([isinstance(call, TextContent) for call in response.calls])
        if is_only_text:
            role = "assistant"
        else:
            role = "function_call"

        return Message(
            role=role,  # global role
            content=[
                ToolCall(
                    args=result.args,
                    name=result.name,
                    call_id=result.call_id,
                )
                if isinstance(result, ToolCallResponse)
                else result
                for result in response.calls
            ],
        )



import os
from typing import Mapping, Union

import httpx
from anthropic import Anthropic, AsyncAnthropic
from anthropic._base_client import DEFAULT_MAX_RETRIES
from anthropic._types import (
    NOT_GIVEN,
    NotGiven,
    ProxiesTypes,
    Timeout,
    Transport,
)


def get_anthropic_client(
    api_key_env_name: str | None = None,
    auth_token_env_name: str | None = None,
    endpoint_env_name: str | httpx.URL | None = None,
    timeout: Union[float, Timeout, None, NotGiven] = NOT_GIVEN,
    max_retries: int = DEFAULT_MAX_RETRIES,
    default_headers: Mapping[str, str] | None = None,
    default_query: Mapping[str, object] | None = None,
    http_client: httpx.Client | None = None,
    transport: Transport | None = None,
    proxies: ProxiesTypes | None = None,
    connection_pool_limits: httpx.Limits | None = None,
    _strict_response_validation: bool = False,
) -> Anthropic:
    return Anthropic(
        api_key=os.getenv(api_key_env_name) if api_key_env_name else None,
        auth_token=os.getenv(auth_token_env_name) if auth_token_env_name else None,
        base_url=os.getenv(endpoint_env_name) if endpoint_env_name else None,
        timeout=timeout,
        max_retries=max_retries,
        default_headers=default_headers,
        default_query=default_query,
        http_client=http_client,
        transport=transport,
        proxies=proxies,
        connection_pool_limits=connection_pool_limits,
        _strict_response_validation=_strict_response_validation,
    )


def get_async_anthropic_client(
    api_key_env_name: str | None = None,
    auth_token_env_name: str | None = None,
    endpoint_env_name: str | httpx.URL | None = None,
    timeout: Union[float, Timeout, None, NotGiven] = NOT_GIVEN,
    max_retries: int = DEFAULT_MAX_RETRIES,
    default_headers: Mapping[str, str] | None = None,
    default_query: Mapping[str, object] | None = None,
    http_client: httpx.Client | None = None,
    transport: Transport | None = None,
    proxies: ProxiesTypes | None = None,
    connection_pool_limits: httpx.Limits | None = None,
    _strict_response_validation: bool = False,
) -> AsyncAnthropic:
    return AsyncAnthropic(
        api_key=os.getenv(api_key_env_name) if api_key_env_name else None,
        auth_token=os.getenv(auth_token_env_name) if auth_token_env_name else None,
        base_url=os.getenv(endpoint_env_name) if endpoint_env_name else None,
        timeout=timeout,
        max_retries=max_retries,
        default_headers=default_headers,
        default_query=default_query,
        http_client=http_client,
        transport=transport,
        proxies=proxies,
        connection_pool_limits=connection_pool_limits,
        _strict_response_validation=_strict_response_validation,
    )



import os
from typing import (
    Any,
    Awaitable,
    Callable,
    Mapping,
    Sequence,
    Union,
)

from qdrant_client import AsyncQdrantClient, QdrantClient, models
from qdrant_client.conversions import common_types as types

from ..result import RetrievalResults
from .base import (
    BaseEmbeddingModule,
    BaseLocalCollectionModule,
    BaseLocalRetrievalModule,
    BaseRemoteCollectionModule,
    BaseRemoteRetrievalModule,
)


class QdrantLocalCollectionModule(BaseLocalCollectionModule):
    def __init__(
        self,
        persistence_directory: str,
        collection_name: str,
        vectors_config: Union[types.VectorParams, Mapping[str, types.VectorParams]],
        embedder: BaseEmbeddingModule | None = None,
        logger: Any | None = None,
        sparse_vectors_config: Mapping[str, types.SparseVectorParams] | None = None,
        on_disk_payload: bool | None = None,
        hnsw_config: types.HnswConfigDiff | None = None,
        optimizers_config: types.OptimizersConfigDiff | None = None,
        wal_config: types.WalConfigDiff | None = None,
        quantization_config: types.QuantizationConfig | None = None,
        init_from: types.InitFrom | None = None,
        timeout: int | None = None,
        force_disable_check_same_thread: bool = False,
    ):
        super().__init__(
            persistence_directory=persistence_directory,
            collection_name=collection_name,
            embedder=embedder,
            logger=logger,
        )
        self.vectors_config = vectors_config
        self.sparse_vectors_config = sparse_vectors_config
        self.on_disk_payload = on_disk_payload
        self.hnsw_config = hnsw_config
        self.optimizers_config = optimizers_config
        self.wal_config = wal_config
        self.quantization_config = quantization_config
        self.init_from = init_from
        self.timeout = timeout
        self.force_disable_check_same_thread = force_disable_check_same_thread

    def _exists(self, client: QdrantClient, collection_name: str) -> bool:
        return client.collection_exists(collection_name=collection_name)

    def _create_collection(self, client: QdrantClient, collection_name: str) -> None:
        client.create_collection(
            collection_name=collection_name,
            vectors_config=self.vectors_config,
            sparse_vectors_config=self.sparse_vectors_config,
            on_disk_payload=self.on_disk_payload,
            hnsw_config=self.hnsw_config,
            optimizers_config=self.optimizers_config,
            wal_config=self.wal_config,
            quantization_config=self.quantization_config,
            init_from=self.init_from,
            timeout=self.timeout,
        )

    def _upsert(
        self,
        client: QdrantClient,
        collection_name: str,
        ids: list[str | int],
        documents: list[str],
        embeddings: list[list[float]],
        metadatas: list[dict[str, str]],
        wait: bool = True,
        ordering: types.WriteOrdering | None = None,
        shard_key_selector: types.ShardKeySelector | None = None,
    ) -> None:
        client.upsert(
            collection_name=collection_name,
            points=models.Batch(
                ids=ids,
                vectors=embeddings,
                payloads=[
                    m | {"document": doc} for m, doc in zip(metadatas, documents, strict=True)
                ],
            ),
            wait=wait,
            ordering=ordering,
            shard_key_selector=shard_key_selector,
        )

        return

    def _delete_collection(self, client: QdrantClient, collection_name: str) -> None:
        client.delete_collection(collection_name=collection_name)

    def _delete_record(
        self,
        client: QdrantClient,
        collection_name: str,
        points_selector: types.PointsSelector,
        wait: bool = True,
        ordering: types.WriteOrdering | None = None,
        shard_key_selector: types.ShardKeySelector | None = None,
    ) -> None:
        client.delete(
            collection_name=collection_name,
            points_selector=points_selector,
            wait=wait,
            ordering=ordering,
            shard_key_selector=shard_key_selector,
        )

    def get_client(self) -> QdrantClient:
        return QdrantClient(
            path=self.persistence_directory,
            force_disable_check_same_thread=self.force_disable_check_same_thread,
        )

    def as_retriever(
        self,
        n_results: int = 4,
        score_threshold: float = 0.5,
        ascending: bool = False,
    ) -> "QdrantLocalRetrievalModule":
        return QdrantLocalRetrievalModule(
            embedder=self.embedder,
            persistence_directory=self.persistence_directory,
            collection_name=self.collection_name,
            n_results=n_results,
            score_threshold=score_threshold,
            logger=self.logger,
            force_disable_check_same_thread=self.force_disable_check_same_thread,
            ascending=ascending,
        )


class QdrantRemoteCollectionModule(BaseRemoteCollectionModule):
    def __init__(
        self,
        url: str,
        collection_name: str,
        port: str = "6333",
        embedder: BaseEmbeddingModule | None = None,
        logger: Any | None = None,
        https: bool | None = None,
        api_key_env_name: str | None = None,
        host: str | None = None,
        grpc_port: str | None = "6334",
        grpc_options: dict[str, Any] | None = None,
        auth_token_provider: Union[Callable[[], str], Callable[[], Awaitable[str]]] | None = None,
        vectors_config: Union[types.VectorParams, Mapping[str, types.VectorParams]] = None,
        sparse_vectors_config: Mapping[str, types.SparseVectorParams] | None = None,
        shard_number: int | None = None,
        sharding_method: types.ShardingMethod | None = None,
        replication_factor: int | None = None,
        write_consistency_factor: int | None = None,
        on_disk_payload: bool | None = None,
        hnsw_config: types.HnswConfigDiff | None = None,
        optimizers_config: types.OptimizersConfigDiff | None = None,
        wal_config: types.WalConfigDiff | None = None,
        quantization_config: types.QuantizationConfig | None = None,
        init_from: types.InitFrom | None = None,
        timeout: int | None = None,
    ):
        super().__init__(
            collection_name=collection_name,
            embedder=embedder,
            logger=logger,
            url=url,
            port=port,
        )
        self.https = https
        self.api_key_env_name = api_key_env_name
        self.host = host
        self.grpc_port = grpc_port
        self.grpc_options = grpc_options
        self.auth_token_provider = auth_token_provider
        self.vectors_config = vectors_config
        self.sparse_vectors_config = sparse_vectors_config
        self.shard_number = shard_number
        self.sharding_method = sharding_method
        self.replication_factor = replication_factor
        self.write_consistency_factor = write_consistency_factor
        self.on_disk_payload = on_disk_payload
        self.hnsw_config = hnsw_config
        self.optimizers_config = optimizers_config
        self.wal_config = wal_config
        self.quantization_config = quantization_config
        self.init_from = init_from
        self.timeout = timeout

    def _exists(self, client: QdrantClient, collection_name: str) -> bool:
        return client.collection_exists(collection_name=collection_name)

    def _create_collection(self, client: QdrantClient, collection_name: str) -> None:
        client.create_collection(
            collection_name=collection_name,
            vectors_config=self.vectors_config,
            sparse_vectors_config=self.sparse_vectors_config,
            shard_number=self.shard_number,
            sharding_method=self.sharding_method,
            replication_factor=self.replication_factor,
            write_consistency_factor=self.write_consistency_factor,
            on_disk_payload=self.on_disk_payload,
            hnsw_config=self.hnsw_config,
            optimizers_config=self.optimizers_config,
            wal_config=self.wal_config,
            quantization_config=self.quantization_config,
            init_from=self.init_from,
            timeout=self.timeout,
        )

    def _upsert(
        self,
        client: QdrantClient,
        collection_name: str,
        ids: list[str | int],
        documents: list[str],
        embeddings: list[list[float]],
        metadatas: list[dict[str, str]],
        wait: bool = True,
        ordering: types.WriteOrdering | None = None,
        shard_key_selector: types.ShardKeySelector | None = None,
    ) -> None:
        client.upsert(
            collection_name=collection_name,
            points=models.Batch(
                ids=ids,
                vectors=embeddings,
                payloads=[
                    m | {"document": doc} for m, doc in zip(metadatas, documents, strict=True)
                ],
            ),
            wait=wait,
            ordering=ordering,
            shard_key_selector=shard_key_selector,
        )

        return

    def _delete_collection(self, client: QdrantClient, collection_name: str) -> None:
        client.delete_collection(collection_name=collection_name)

    def _delete_record(
        self,
        client: QdrantClient,
        collection_name: str,
        points_selector: types.PointsSelector,
        wait: bool = True,
        ordering: types.WriteOrdering | None = None,
        shard_key_selector: types.ShardKeySelector | None = None,
    ) -> None:
        client.delete(
            collection_name=collection_name,
            points_selector=points_selector,
            wait=wait,
            ordering=ordering,
            shard_key_selector=shard_key_selector,
        )

    async def _aexists(self, client: AsyncQdrantClient, collection_name: str) -> bool:
        return await client.collection_exists(collection_name=collection_name)

    async def _acreate_collection(self, client: AsyncQdrantClient, collection_name: str) -> None:
        await client.create_collection(
            collection_name=collection_name,
            vectors_config=self.vectors_config,
            sparse_vectors_config=self.sparse_vectors_config,
            shard_number=self.shard_number,
            sharding_method=self.sharding_method,
            replication_factor=self.replication_factor,
            write_consistency_factor=self.write_consistency_factor,
            on_disk_payload=self.on_disk_payload,
            hnsw_config=self.hnsw_config,
            optimizers_config=self.optimizers_config,
            wal_config=self.wal_config,
            quantization_config=self.quantization_config,
            init_from=self.init_from,
            timeout=self.timeout,
        )

    async def _aupsert(
        self,
        client: AsyncQdrantClient,
        collection_name: str,
        ids: list[str | int],
        documents: list[str],
        embeddings: list[list[float]],
        metadatas: list[dict[str, str]],
        wait: bool = True,
        ordering: types.WriteOrdering | None = None,
        shard_key_selector: types.ShardKeySelector | None = None,
    ) -> None:
        await client.upsert(
            collection_name=collection_name,
            points=models.Batch(
                ids=ids,
                vectors=embeddings,
                payloads=[
                    m | {"document": doc} for m, doc in zip(metadatas, documents, strict=True)
                ],
            ),
            wait=wait,
            ordering=ordering,
            shard_key_selector=shard_key_selector,
        )

        return

    async def _adelete_collection(self, client: AsyncQdrantClient, collection_name: str) -> None:
        await client.delete_collection(collection_name=collection_name)

    async def _adelete_record(
        self,
        client: AsyncQdrantClient,
        collection_name: str,
        points_selector: types.PointsSelector,
        wait: bool = True,
        ordering: types.WriteOrdering | None = None,
        shard_key_selector: types.ShardKeySelector | None = None,
    ) -> None:
        await client.delete(
            collection_name=collection_name,
            points_selector=points_selector,
            wait=wait,
            ordering=ordering,
            shard_key_selector=shard_key_selector,
        )

    def get_client(self) -> QdrantClient:
        if hasattr(self, "client") and isinstance(self.client, QdrantClient):
            return self.client

        self.client = QdrantClient(
            url=self.url,
            port=self.port,
            https=self.https,
            api_key=os.getenv(self.api_key_env_name) if self.api_key_env_name else None,
            host=self.host,
            grpc_port=self.grpc_port,
            grpc_options=self.grpc_options,
            auth_token_provider=self.auth_token_provider,
            timeout=self.timeout,
        )
        return self.client

    def get_async_client(self) -> AsyncQdrantClient:
        if hasattr(self, "client") and isinstance(self.client, AsyncQdrantClient):
            return self.client

        self.client = AsyncQdrantClient(
            url=self.url,
            port=self.port,
            https=self.https,
            api_key=os.getenv(self.api_key_env_name) if self.api_key_env_name else None,
            host=self.host,
            grpc_port=self.grpc_port,
            grpc_options=self.grpc_options,
            auth_token_provider=self.auth_token_provider,
            timeout=self.timeout,
        )
        return self.client

    def as_retriever(
        self,
        n_results: int = 4,
        score_threshold: float = 0.5,
        ascending: bool = False,
    ) -> "QdrantRemoteRetrievalModule":
        return QdrantRemoteRetrievalModule(
            embedder=self.embedder,
            url=self.url,
            port=self.port,
            collection_name=self.collection_name,
            n_results=n_results,
            score_threshold=score_threshold,
            logger=self.logger,
            https=self.https,
            api_key_env_name=self.api_key_env_name,
            host=self.host,
            grpc_port=self.grpc_port,
            grpc_options=self.grpc_options,
            auth_token_provider=self.auth_token_provider,
            timeout=self.timeout,
            ascending=ascending,
        )


class QdrantLocalRetrievalModule(BaseLocalRetrievalModule):
    def __init__(
        self,
        persistence_directory: str,
        collection_name: str,
        embedder: BaseEmbeddingModule | None = None,
        n_results: int = 4,
        score_threshold: float = 0.5,
        logger: Any | None = None,
        ascending: bool = False,
        force_disable_check_same_thread: bool = False,
    ):
        super().__init__(
            persistence_directory=persistence_directory,
            collection_name=collection_name,
            embedder=embedder,
            n_results=n_results,
            score_threshold=score_threshold,
            logger=logger,
            ascending=ascending,
        )
        self.force_disable_check_same_thread = force_disable_check_same_thread

    def get_client(self) -> QdrantClient:
        return QdrantClient(
            path=self.persistence_directory,
            force_disable_check_same_thread=self.force_disable_check_same_thread,
        )

    def _retrieve(
        self,
        client: QdrantClient,
        collection_name: str,
        query_vector: list[float],
        n_results: int,
        score_threshold: float,
        filter: Any | None = None,
        search_params: types.SearchParams | None = None,
        offset: int | None = None,
        with_payload: Union[bool, Sequence[str], types.PayloadSelector] = True,
        with_vectors: Union[bool, Sequence[str]] = False,
        append_payload: bool = True,
        consistency: types.ReadConsistency | None = None,
        shard_key_selector: types.ShardKeySelector | None = None,
        timeout: int | None = None,
    ) -> RetrievalResults:
        retrieved = client.search(
            collection_name=collection_name,
            query_vector=query_vector,
            query_filter=filter,
            score_threshold=score_threshold,
            limit=n_results,
            search_params=search_params,
            offset=offset,
            with_payload=with_payload,
            with_vectors=with_vectors,
            append_payload=append_payload,
            consistency=consistency,
            shard_key_selector=shard_key_selector,
            timeout=timeout,
        )

        ids = [r.id for r in retrieved]
        scores = [r.score for r in retrieved]
        documents = [r.payload["document"] for r in retrieved]
        metadatas = [r.payload for r in retrieved]
        collections = [collection_name for _ in retrieved]

        return RetrievalResults(
            ids=ids,
            scores=scores,
            documents=documents,
            metadatas=metadatas,
            collections=collections,
        )


class QdrantRemoteRetrievalModule(BaseRemoteRetrievalModule):
    def __init__(
        self,
        url: str,
        collection_name: str,
        port: str = "6333",
        embedder: BaseEmbeddingModule | None = None,
        n_results: int = 4,
        score_threshold: float = 0.5,
        logger: Any | None = None,
        ascending: bool = False,
        https: bool | None = None,
        api_key_env_name: str | None = None,
        host: str | None = None,
        grpc_port: str | None = "6334",
        grpc_options: dict[str, Any] | None = None,
        auth_token_provider: Union[Callable[[], str], Callable[[], Awaitable[str]]] | None = None,
        timeout: int | None = None,
    ):
        super().__init__(
            collection_name=collection_name,
            embedder=embedder,
            n_results=n_results,
            score_threshold=score_threshold,
            logger=logger,
            url=url,
            port=port,
            ascending=ascending,
        )
        self.https = https
        self.api_key_env_name = api_key_env_name
        self.host = host
        self.grpc_port = grpc_port
        self.grpc_options = grpc_options
        self.auth_token_provider = auth_token_provider
        self.timeout = timeout

    def get_client(self) -> QdrantClient:
        if hasattr(self, "client") and isinstance(self.client, QdrantClient):
            return self.client

        self.client = QdrantClient(
            url=self.url,
            port=self.port,
            https=self.https,
            api_key=os.getenv(self.api_key_env_name) if self.api_key_env_name else None,
            host=self.host,
            grpc_port=self.grpc_port,
            grpc_options=self.grpc_options,
            auth_token_provider=self.auth_token_provider,
            timeout=self.timeout,
        )
        return self.client

    def get_async_client(self) -> AsyncQdrantClient:
        if hasattr(self, "client") and isinstance(self.client, AsyncQdrantClient):
            return self.client

        self.client = AsyncQdrantClient(
            url=self.url,
            port=self.port,
            https=self.https,
            api_key=os.getenv(self.api_key_env_name) if self.api_key_env_name else None,
            host=self.host,
            grpc_port=self.grpc_port,
            grpc_options=self.grpc_options,
            auth_token_provider=self.auth_token_provider,
            timeout=self.timeout,
        )
        return self.client

    def _retrieve(
        self,
        client: QdrantClient,
        collection_name: str,
        query_vector: list[float],
        n_results: int,
        score_threshold: float,
        filter: Any | None = None,
        search_params: types.SearchParams | None = None,
        offset: int | None = None,
        with_payload: Union[bool, Sequence[str], types.PayloadSelector] = True,
        with_vectors: Union[bool, Sequence[str]] = False,
        append_payload: bool = True,
        consistency: types.ReadConsistency | None = None,
        shard_key_selector: types.ShardKeySelector | None = None,
        timeout: int | None = None,
    ) -> RetrievalResults:
        retrieved = client.search(
            collection_name=collection_name,
            query_vector=query_vector,
            query_filter=filter,
            score_threshold=score_threshold,
            limit=n_results,
            search_params=search_params,
            offset=offset,
            with_payload=with_payload,
            with_vectors=with_vectors,
            append_payload=append_payload,
            consistency=consistency,
            shard_key_selector=shard_key_selector,
            timeout=timeout,
        )

        ids = [r.id for r in retrieved]
        scores = [r.score for r in retrieved]
        documents = [r.payload["document"] for r in retrieved]
        metadatas = [r.payload for r in retrieved]
        collections = [collection_name for _ in retrieved]

        return RetrievalResults(
            ids=ids,
            scores=scores,
            documents=documents,
            metadatas=metadatas,
            collections=collections,
        )

    async def _aretrieve(
        self,
        client: AsyncQdrantClient,
        collection_name: str,
        query_vector: list[float],
        n_results: int,
        score_threshold: float,
        filter: Any | None = None,
        **kwargs,
    ) -> RetrievalResults:
        retrieved = await client.search(
            collection_name=collection_name,
            query_vector=query_vector,
            query_filter=filter,
            score_threshold=score_threshold,
            with_vectors=False,
            limit=n_results,
            **kwargs,
        )

        ids = [r.id for r in retrieved]
        scores = [r.score for r in retrieved]
        documents = [r.payload["document"] for r in retrieved]
        metadatas = [r.payload for r in retrieved]
        collections = [collection_name for _ in retrieved]

        return RetrievalResults(
            ids=ids,
            scores=scores,
            documents=documents,
            metadatas=metadatas,
            collections=collections,
        )



from typing import Any

import chromadb
from chromadb import DEFAULT_DATABASE, DEFAULT_TENANT, Settings
from chromadb.api import ClientAPI
from chromadb.api.types import URI, Image, Include, OneOrMany
from chromadb.types import Where, WhereDocument

from ..result import RetrievalResults
from .base import (
    BaseEmbeddingModule,
    BaseLocalCollectionModule,
    BaseLocalRetrievalModule,
    BaseRemoteCollectionModule,
    BaseRemoteRetrievalModule,
)


class ChromaLocalCollectionModule(BaseLocalCollectionModule):
    def __init__(
        self,
        persistence_directory: str,
        collection_name: str,
        metadata: dict[str, Any] | None = None,
        embedder: BaseEmbeddingModule | None = None,
        logger: Any | None = None,
        settings: Settings | None = None,
        tenant: str = DEFAULT_TENANT,
        database: str = DEFAULT_DATABASE,
    ):
        super().__init__(
            persistence_directory=persistence_directory,
            collection_name=collection_name,
            embedder=embedder,
            logger=logger,
        )
        self.metadata = metadata or {"hnsw:space": "cosine"}
        self.settings = settings
        self.tenant = tenant
        self.database = database

    def _exists(self, client: ClientAPI, collection_name: str) -> bool:
        return len([c.name for c in client.list_collections() if c.name == collection_name]) > 0

    def _create_collection(self, client: ClientAPI, collection_name: str) -> None:
        self.collection = client.create_collection(name=collection_name, metadata=self.metadata)

    def _delete_record(
        self,
        client: ClientAPI,
        collection_name: str,
        ids: list[str | int],
        filter: Where | None = None,
        where_document: WhereDocument | None = None,
    ) -> None:
        if not hasattr(self, "collection"):
            self.collection = client.get_collection(name=collection_name)
        self.collection.delete(
            ids=[str(i) for i in ids], where=filter, where_document=where_document
        )

    def _upsert(
        self,
        client: ClientAPI,
        collection_name: str,
        ids: list[str | int],
        documents: list[str],
        embeddings: list[list[float]],
        metadatas: list[dict[str, str]],
        images: OneOrMany[Image] | None = None,
        uris: OneOrMany[URI] | None = None,
    ) -> None:
        if not hasattr(self, "collection"):
            self.collection = client.get_collection(name=collection_name)

        self.collection.upsert(
            ids=[str(i) for i in ids],
            embeddings=embeddings,
            documents=documents,
            metadatas=[m | {"document": doc} for m, doc in zip(metadatas, documents, strict=True)],
            images=images,
            uris=uris,
        )

        return

    def _delete_collection(self, client: ClientAPI, collection_name: str) -> None:
        client.delete_collection(name=collection_name)

    def get_client(self) -> ClientAPI:
        return chromadb.PersistentClient(
            path=self.persistence_directory.as_posix(),
            settings=self.settings,
            tenant=self.tenant,
            database=self.database,
        )

    def as_retriever(
        self,
        n_results: int = 4,
        score_threshold: float = 0.5,
    ) -> "ChromaLocalRetrievalModule":
        return ChromaLocalRetrievalModule(
            embedder=self.embedder,
            persistence_directory=self.persistence_directory,
            collection_name=self.collection_name,
            settings=self.settings,
            tenant=self.tenant,
            database=self.database,
            n_results=n_results,
            score_threshold=score_threshold,
            logger=self.logger,
        )


class ChromaRemoteCollectionModule(BaseRemoteCollectionModule):
    def __init__(
        self,
        host: str,
        collection_name: str,
        port: str = "8000",
        ssl: bool = False,
        headers: dict[str, str] | None = None,
        settings: Settings | None = None,
        metadata: dict[str, Any] | None = None,
        embedder: BaseEmbeddingModule | None = None,
        logger: Any | None = None,
        tenant: str = DEFAULT_TENANT,
        database: str = DEFAULT_DATABASE,
    ):
        super().__init__(
            url=host,
            port=port,
            collection_name=collection_name,
            embedder=embedder,
            logger=logger,
        )
        self.metadata = metadata or {"hnsw:space": "cosine"}
        self.tenant = tenant
        self.database = database
        self.ssl = ssl
        self.headers = headers
        self.settings = settings

    def _exists(self, client: ClientAPI, collection_name: str) -> bool:
        return len([c.name for c in client.list_collections() if c.name == collection_name]) > 0

    def _create_collection(self, client: ClientAPI, collection_name: str) -> None:
        self.collection = client.create_collection(name=collection_name, metadata=self.metadata)

    def _delete_record(
        self,
        client: ClientAPI,
        collection_name: str,
        ids: list[str | int],
        filter: Where | None = None,
        where_document: WhereDocument | None = None,
    ) -> None:
        if not hasattr(self, "collection"):
            self.collection = client.get_collection(name=collection_name)
        self.collection.delete(
            ids=[str(i) for i in ids], where=filter, where_document=where_document
        )

    def _upsert(
        self,
        client: ClientAPI,
        collection_name: str,
        ids: list[str | int],
        documents: list[str],
        embeddings: list[list[float]],
        metadatas: list[dict[str, str]],
        images: OneOrMany[Image] | None = None,
        uris: OneOrMany[URI] | None = None,
    ) -> None:
        if not hasattr(self, "collection"):
            self.collection = client.get_collection(name=collection_name)

        self.collection.upsert(
            ids=[str(i) for i in ids],
            embeddings=embeddings,
            documents=documents,
            metadatas=[m | {"document": doc} for m, doc in zip(metadatas, documents, strict=True)],
            images=images,
            uris=uris,
        )

        return

    def _delete_collection(self, client: ClientAPI, collection_name: str) -> None:
        client.delete_collection(name=collection_name)

    def get_client(self) -> ClientAPI:
        return chromadb.HttpClient(
            host=self.url,
            port=self.port,
            ssl=self.ssl,
            headers=self.headers,
            settings=self.settings,
            tenant=self.tenant,
            database=self.database,
        )

    def get_async_client(self) -> ClientAPI:
        return self.get_client()

    def as_retriever(
        self,
        n_results: int = 4,
        score_threshold: float = 0.5,
    ) -> "ChromaRemoteRetrievalModule":
        return ChromaRemoteRetrievalModule(
            embedder=self.embedder,
            host=self.url,
            port=self.port,
            ssl=self.ssl,
            headers=self.headers,
            settings=self.settings,
            collection_name=self.collection_name,
            n_results=n_results,
            score_threshold=score_threshold,
            logger=self.logger,
            tenant=self.tenant,
            database=self.database,
        )

    async def _acreate_collection(self, client: ClientAPI, collection_name: str) -> None:
        self._create_collection(client=client, collection_name=collection_name)

    async def _aexists(self, client: Any, collection_name: str) -> bool:
        return self._exists(client=client, collection_name=collection_name)

    async def _adelete_collection(self, client: ClientAPI, collection_name: str) -> None:
        self._delete_collection(client=client, collection_name=collection_name)

    async def _adelete_record(
        self,
        client: ClientAPI,
        collection_name: str,
        ids: list[str | int],
        filter: Where | None = None,
        where_document: WhereDocument | None = None,
    ) -> None:
        self._delete_record(
            client=client,
            collection_name=collection_name,
            ids=ids,
            filter=filter,
            where_document=where_document,
        )

    async def _aupsert(
        self,
        client: ClientAPI,
        collection_name: str,
        ids: list[str | int],
        documents: list[str],
        embeddings: list[list[float]],
        metadatas: list[dict[str, str]],
        images: OneOrMany[Image] | None = None,
        uris: OneOrMany[URI] | None = None,
    ) -> None:
        self._upsert(
            client=client,
            collection_name=collection_name,
            ids=ids,
            documents=documents,
            embeddings=embeddings,
            metadatas=metadatas,
            images=images,
            uris=uris,
        )


class ChromaLocalRetrievalModule(BaseLocalRetrievalModule):
    def __init__(
        self,
        persistence_directory: str,
        collection_name: str,
        embedder: BaseEmbeddingModule | None = None,
        n_results: int = 4,
        score_threshold: float = 0.5,
        logger: Any | None = None,
        ascending: bool = True,
        settings: Settings | None = None,
        tenant: str = DEFAULT_TENANT,
        database: str = DEFAULT_DATABASE,
    ):
        super().__init__(
            persistence_directory=persistence_directory,
            collection_name=collection_name,
            embedder=embedder,
            n_results=n_results,
            score_threshold=score_threshold,
            logger=logger,
            ascending=ascending,
        )
        self.settings = settings
        self.tenant = tenant
        self.database = database

    def get_client(self) -> ClientAPI:
        return chromadb.PersistentClient(
            path=self.persistence_directory.as_posix(),
            settings=self.settings,
            tenant=self.tenant,
            database=self.database,
        )

    def _retrieve(
        self,
        client: ClientAPI,
        collection_name: str,
        query_vector: list[float],
        n_results: int,
        score_threshold: float,
        filter: Any | None = None,
        query_images: OneOrMany[Image] | None = None,
        query_uris: OneOrMany[URI] | None = None,
        where_document: WhereDocument | None = None,
        include: Include | None = None,
    ) -> RetrievalResults:
        if not hasattr(self, "collection"):
            self.collection = client.get_collection(name=collection_name)

        retrieved = self.collection.query(
            query_embeddings=query_vector,
            where=filter,
            include=include or ["metadatas", "documents", "distances"],
            n_results=n_results,
            query_images=query_images,
            query_uris=query_uris,
            where_document=where_document,
        )

        ids = []
        scores = []
        documents = []
        metadatas = []
        collections = []

        for _id, dist, document, metadata in zip(
            retrieved["ids"][0],
            retrieved["distances"][0],
            retrieved["documents"][0],
            retrieved["metadatas"][0],
            strict=True,
        ):
            if dist <= score_threshold:
                ids.append(int(_id))
                scores.append(dist)
                documents.append(document)
                metadatas.append(metadata)
                collections.append(collection_name)

        return RetrievalResults(
            ids=ids,
            scores=scores,
            documents=documents,
            metadatas=metadatas,
            collections=collections,
        )


class ChromaRemoteRetrievalModule(BaseRemoteRetrievalModule):
    def __init__(
        self,
        host: str,
        collection_name: str,
        port: str = "8000",
        ssl: bool = False,
        headers: dict[str, str] | None = None,
        settings: Settings | None = None,
        embedder: BaseEmbeddingModule | None = None,
        n_results: int = 4,
        score_threshold: float = 0.5,
        logger: Any | None = None,
        ascending: bool = True,
        tenant: str = DEFAULT_TENANT,
        database: str = DEFAULT_DATABASE,
    ):
        super().__init__(
            url=host,
            port=port,
            collection_name=collection_name,
            embedder=embedder,
            n_results=n_results,
            score_threshold=score_threshold,
            logger=logger,
            ascending=ascending,
        )
        self.ssl = ssl
        self.headers = headers
        self.settings = settings
        self.tenant = tenant
        self.database = database

    def get_client(self) -> ClientAPI:
        return chromadb.HttpClient(
            host=self.url,
            port=self.port,
            ssl=self.ssl,
            headers=self.headers,
            settings=self.settings,
            tenant=self.tenant,
            database=self.database,
        )

    def get_async_client(self) -> ClientAPI:
        return self.get_client()

    def _retrieve(
        self,
        client: ClientAPI,
        collection_name: str,
        query_vector: list[float],
        n_results: int,
        score_threshold: float,
        filter: Any | None = None,
        query_images: OneOrMany[Image] | None = None,
        query_uris: OneOrMany[URI] | None = None,
        where_document: WhereDocument | None = None,
        include: Include | None = None,
    ) -> RetrievalResults:
        if not hasattr(self, "collection"):
            self.collection = client.get_collection(name=collection_name)

        retrieved = self.collection.query(
            query_embeddings=query_vector,
            where=filter,
            include=include or ["metadatas", "documents", "distances"],
            n_results=n_results,
            query_images=query_images,
            query_uris=query_uris,
            where_document=where_document,
        )

        ids = []
        scores = []
        documents = []
        metadatas = []
        collections = []

        for _id, dist, document, metadata in zip(
            retrieved["ids"][0],
            retrieved["distances"][0],
            retrieved["documents"][0],
            retrieved["metadatas"][0],
            strict=True,
        ):
            if dist <= score_threshold:
                ids.append(int(_id))
                scores.append(dist)
                documents.append(document)
                metadatas.append(metadata)
                collections.append(collection_name)

        return RetrievalResults(
            ids=ids,
            scores=scores,
            documents=documents,
            metadatas=metadatas,
            collections=collections,
        )

    async def _aretrieve(
        self,
        client: ClientAPI,
        collection_name: str,
        query_vector: list[float],
        n_results: int,
        score_threshold: float,
        filter: Any | None = None,
        query_images: OneOrMany[Image] | None = None,
        query_uris: OneOrMany[URI] | None = None,
        where_document: WhereDocument | None = None,
        include: Include | None = None,
    ) -> RetrievalResults:
        return self._retrieve(
            client=client,
            collection_name=collection_name,
            query_vector=query_vector,
            n_results=n_results,
            score_threshold=score_threshold,
            filter=filter,
            query_images=query_images,
            query_uris=query_uris,
            where_document=where_document,
            include=include,
        )



import asyncio
import math
import time
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Optional

from tqdm import tqdm

from langrila import Usage

from ..base import BaseEmbeddingModule
from ..logger import DefaultLogger
from ..result import EmbeddingResults, RetrievalResults
from ..utils import make_batch


class AbstractLocalCollectionModule(ABC):
    """
    Base class for collection module.
    """

    @abstractmethod
    def get_client(self) -> Any:
        """
        return the client object
        """
        raise NotImplementedError

    @abstractmethod
    def _create_collection(self, client: Any, collection_name: str) -> None:
        """
        create a collection
        """
        raise NotImplementedError

    @abstractmethod
    def _exists(self, client: Any, collection_name: str) -> bool:
        """
        check if the collection exists
        """
        raise NotImplementedError

    @abstractmethod
    def _upsert(
        self,
        client: Any,
        collection_name: str,
        ids: list[str | int],
        documents: list[str],
        embeddings: list[list[float]],
        metadatas: list[dict[str, str]],
        **kwargs,
    ) -> None:
        """
        upsert embeddings, documents and metadatas.
        """
        raise NotImplementedError

    @abstractmethod
    def _delete_collection(self, client: Any, collection_name: str) -> None:
        """
        delete the collection
        """
        raise NotImplementedError

    @abstractmethod
    def _delete_record(self, client: Any, collection_name: str, **kwargs) -> None:
        """
        delete the record from the collection. kwargs depends on the clieint.
        """
        raise NotImplementedError

    @abstractmethod
    def as_retriever(
        self,
        n_results: int,
        score_threshold: float,
    ) -> "BaseLocalRetrievalModule":
        """
        return the retrieval module
        """
        raise NotImplementedError


class AbstractRemoteCollectionModule(ABC):
    """
    Base class for collection module.
    """

    @abstractmethod
    def get_client(self) -> Any:
        """
        return the client object
        """
        raise NotImplementedError

    @abstractmethod
    def get_async_client(self) -> Any:
        """
        return the async client object
        """
        raise NotImplementedError

    @abstractmethod
    def _create_collection(self, client: Any, collection_name: str) -> None:
        """
        create a collection
        """
        raise NotImplementedError

    @abstractmethod
    async def _acreate_collection(self, client: Any, collection_name: str) -> None:
        """
        create a collection
        """
        raise NotImplementedError

    @abstractmethod
    def _exists(self, client: Any, collection_name: str) -> bool:
        """
        check if the collection exists
        """
        raise NotImplementedError

    @abstractmethod
    async def _aexists(self, client: Any, collection_name: str) -> bool:
        """
        check if the collection exists
        """
        raise NotImplementedError

    @abstractmethod
    def _upsert(
        self,
        client: Any,
        collection_name: str,
        ids: list[str | int],
        documents: list[str],
        embeddings: list[list[float]],
        metadatas: list[dict[str, str]],
        **kwargs,
    ) -> None:
        """
        upsert embeddings, documents and metadatas.
        """
        raise NotImplementedError

    @abstractmethod
    async def _aupsert(
        self,
        client: Any,
        collection_name: str,
        ids: list[str | int],
        documents: list[str],
        embeddings: list[list[float]],
        metadatas: list[dict[str, str]],
        **kwargs,
    ) -> None:
        """
        upsert embeddings, documents and metadatas.
        """
        raise NotImplementedError

    @abstractmethod
    async def _adelete_collection(self, client: Any, collection_name: str) -> None:
        """
        delete the collection
        """
        raise NotImplementedError

    @abstractmethod
    def _delete_collection(self, client: Any, collection_name: str) -> None:
        """
        delete the collection
        """
        raise NotImplementedError

    @abstractmethod
    async def _adelete_record(self, client: Any, collection_name: str, **kwargs) -> None:
        """
        delete the record from the collection. kwargs depends on the clieint.
        """
        raise NotImplementedError

    @abstractmethod
    def _delete_record(self, client: Any, collection_name: str, **kwargs) -> None:
        """
        delete the record from the collection. kwargs depends on the clieint.
        """
        raise NotImplementedError

    @abstractmethod
    def as_retriever(
        self,
        n_results: int,
        score_threshold: float,
    ) -> "BaseRemoteRetrievalModule":
        """
        return the retrieval module
        """
        raise NotImplementedError


class BaseLocalCollectionModule(AbstractLocalCollectionModule):
    def __init__(
        self,
        persistence_directory: Path | str,
        collection_name: str,
        embedder: BaseEmbeddingModule | None = None,
        logger: Any | None = None,
    ):
        self.persistence_directory = Path(persistence_directory)
        self.embedder = embedder
        self.collection_name = collection_name
        self.logger = logger or DefaultLogger()

    def create_collection(self) -> None:
        client = self.get_client()
        self._create_collection(client=client, collection_name=self.collection_name)

    def exists(self) -> bool:
        client = self.get_client()
        return self._exists(client=client, collection_name=self.collection_name)

    def upsert(
        self,
        ids: list[str | int],
        documents: list[str],
        embeddings: list[list[float]],
        metadatas: list[dict[str, str]],
        **kwargs,
    ) -> None:
        self._verify_metadata(metadatas)
        client = self.get_client()
        self._upsert(
            client=client,
            collection_name=self.collection_name,
            ids=ids,
            documents=documents,
            embeddings=embeddings,
            metadatas=metadatas,
            **kwargs,
        )

    def delete_collection(self) -> None:
        client = self.get_client()

        if self._exists(client=client, collection_name=self.collection_name):
            self._delete_collection(client=client, collection_name=self.collection_name)

    def delete_record(self, **kwargs) -> None:
        client = self.get_client()
        self._delete_record(client=client, collection_name=self.collection_name, **kwargs)

    def run(
        self,
        documents: list[str],
        metadatas: Optional[list[dict[str, str]]] = None,
        **kwargs,
    ) -> None:
        client = self.get_client()

        if metadatas is not None and len(documents) != len(metadatas):
            raise ValueError(
                "The length of documents and metadatas must be the same. "
                f"Got {len(documents)} documents and {len(metadatas)} metadatas."
            )

        if metadatas is None:
            metadatas = [{} for _ in range(len(documents))]
        else:
            # check if the key 'collection' or 'document' is included in metadatas
            self._verify_metadata(metadatas)

        ids = [i for i in range(len(documents))]

        # batchfy due to memory usage
        batch_size: int = 100

        documents_batch = make_batch(documents, batch_size=batch_size)
        metadatas_batch = make_batch(metadatas, batch_size=batch_size)
        ids_batch = make_batch(ids, batch_size=batch_size)

        n_batches: int = math.ceil(len(documents) / batch_size)

        if not self._exists(client=client, collection_name=self.collection_name):
            self._create_collection(client=client, collection_name=self.collection_name)
            self.logger.info(f"Create collection {self.collection_name}.")

        for doc_batch, metadata_batch, id_batch in tqdm(
            zip(documents_batch, metadatas_batch, ids_batch, strict=True),
            total=n_batches,
        ):
            embedding_batch: EmbeddingResults = self.embedder.run(doc_batch)

            n_retries = 0
            while n_retries < 3:
                try:
                    self._upsert(
                        client=client,
                        collection_name=self.collection_name,
                        ids=id_batch,
                        documents=doc_batch,
                        embeddings=embedding_batch.embeddings,
                        metadatas=metadata_batch,
                        **kwargs,
                    )
                    break
                except Exception as e:
                    self.logger.error(f"Error: {e}")
                    n_retries += 1
                    time.sleep(10)

                    if n_retries == 3:
                        raise e

        if hasattr(self, "_save_on_last"):
            self._save_on_last(client=client)

    def _verify_metadata(self, metadatas: list[dict[str, str]]) -> None:
        """
        check if the key 'document' is included in metadatas
        """
        metadata_keys = [metadata for metadata in metadatas if "document" in metadata]
        if len(metadata_keys) > 0:
            raise ValueError("The key 'document' is reserved. Use another key.")


class BaseRemoteCollectionModule(BaseLocalCollectionModule, AbstractRemoteCollectionModule):
    def __init__(
        self,
        url: str,
        collection_name: str,
        port: str,
        embedder: BaseEmbeddingModule = None,
        logger: Any | None = None,
    ):
        self.url = url
        self.port = port
        self.embedder = embedder
        self.collection_name = collection_name
        self.logger = logger or DefaultLogger()

    async def acreate_collection(self) -> None:
        client = self.get_async_client()
        await self._acreate_collection(client=client, collection_name=self.collection_name)

    async def aexists(self) -> bool:
        client = self.get_async_client()
        return await self._aexists(client=client, collection_name=self.collection_name)

    async def aupsert(
        self,
        ids: list[str | int],
        documents: list[str],
        embeddings: list[list[float]],
        metadatas: list[dict[str, str]],
        **kwargs,
    ) -> None:
        self._verify_metadata(metadatas)
        client = self.get_async_client()
        await self._aupsert(
            client=client,
            collection_name=self.collection_name,
            ids=ids,
            documents=documents,
            embeddings=embeddings,
            metadatas=metadatas,
            **kwargs,
        )

    async def adelete_collection(self) -> None:
        client = self.get_async_client()
        if await self._aexists(client=client, collection_name=self.collection_name):
            await self._adelete_collection(client=client, collection_name=self.collection_name)

    async def adelete_record(self, **kwargs) -> None:
        client = self.get_async_client()
        await self._adelete_record(client=client, collection_name=self.collection_name, **kwargs)

    async def arun(
        self,
        documents: list[str],
        metadatas: Optional[list[dict[str, str]]] = None,
        **kwargs,
    ) -> None:
        client = self.get_async_client()

        if metadatas is not None and len(documents) != len(metadatas):
            raise ValueError(
                "The length of documents and metadatas must be the same. "
                f"Got {len(documents)} documents and {len(metadatas)} metadatas."
            )

        if metadatas is None:
            metadatas = [{} for _ in range(len(documents))]
        else:
            # check if the key 'collection' or 'document' is included in metadatas
            self._verify_metadata(metadatas)

        ids = [i for i in range(len(documents))]

        # batchfy due to memory usage
        batch_size: int = 100

        documents_batch = make_batch(documents, batch_size=batch_size)
        metadatas_batch = make_batch(metadatas, batch_size=batch_size)
        ids_batch = make_batch(ids, batch_size=batch_size)

        n_batches: int = math.ceil(len(documents) / batch_size)

        if not await self._aexists(client=client, collection_name=self.collection_name):
            await self._acreate_collection(client=client, collection_name=self.collection_name)
            self.logger.info(f"Create collection {self.collection_name}.")

        for doc_batch, metadata_batch, id_batch in tqdm(
            zip(documents_batch, metadatas_batch, ids_batch, strict=True),
            total=n_batches,
        ):
            embedding_batch: EmbeddingResults = await self.embedder.arun(doc_batch)

            n_retries = 0
            while n_retries < 3:
                try:
                    await self._aupsert(
                        client=client,
                        collection_name=self.collection_name,
                        ids=id_batch,
                        documents=doc_batch,
                        embeddings=embedding_batch.embeddings,
                        metadatas=metadata_batch,
                        **kwargs,
                    )
                    break
                except Exception as e:
                    self.logger.error(f"Error: {e}")
                    n_retries += 1
                    await asyncio.sleep(10)

                    if n_retries == 3:
                        raise e

        if hasattr(self, "_save_on_last"):
            self._save_on_last(client=client)


class AbstractLocalRetrievalModule(ABC):
    @abstractmethod
    def get_client(self) -> Any:
        """
        return the async client object
        """
        raise NotImplementedError

    @abstractmethod
    def _retrieve(
        self,
        client: Any,
        collection_name: str,
        query_vector: list[float],
        n_results: int,
        score_threshold: float,
        filter: Any | None = None,
        **kwargs,
    ) -> RetrievalResults:
        raise NotImplementedError


class AbstractRemoteRetrievalModule(AbstractLocalRetrievalModule):
    @abstractmethod
    def get_client(self) -> Any:
        raise NotImplementedError

    @abstractmethod
    def get_async_client(self) -> Any:
        """
        return the async client object
        """
        raise NotImplementedError

    @abstractmethod
    def _retrieve(
        self,
        client: Any,
        collection_name: str,
        query_vector: list[float],
        n_results: int,
        score_threshold: float,
        filter: Any | None = None,
        **kwargs,
    ) -> RetrievalResults:
        raise NotImplementedError

    @abstractmethod
    async def _aretrieve(
        self,
        client: Any,
        collection_name: str,
        query_vector: list[float],
        n_results: int,
        score_threshold: float,
        filter: Any | None = None,
        **kwargs,
    ) -> RetrievalResults:
        raise NotImplementedError


class BaseLocalRetrievalModule(AbstractLocalRetrievalModule):
    def __init__(
        self,
        persistence_directory: Path | str,
        collection_name: str,
        embedder: BaseEmbeddingModule = None,
        n_results: int = 4,
        score_threshold: float = 0.5,
        logger: Any | None = None,
        ascending: bool = False,
    ):
        self.persistence_directory = Path(persistence_directory)
        self.embedder = embedder
        self.collection_name = collection_name
        self.n_results = n_results
        self.score_threshold = score_threshold
        self.logger = logger or DefaultLogger()
        self.ascending = ascending

    def run(self, query: str, filter: Any | None = None, **kwargs) -> RetrievalResults:
        client = self.get_client()

        embed: EmbeddingResults = self.embedder.run(query)

        self.logger.info(f"Retrieve from collection {self.collection_name}...")
        retrieved: RetrievalResults = self._retrieve(
            client=client,
            collection_name=self.collection_name,
            query_vector=embed.embeddings[0],
            filter=filter,
            n_results=self.n_results,
            score_threshold=self.score_threshold,
            **kwargs,
        )

        retrieved.usage = embed.usage  # override

        return retrieved


class BaseRemoteRetrievalModule(BaseLocalRetrievalModule, AbstractRemoteRetrievalModule):
    def __init__(
        self,
        url: str,
        collection_name: str,
        port: str,
        embedder: BaseEmbeddingModule = None,
        n_results: int = 4,
        score_threshold: float = 0.5,
        logger: Any | None = None,
        ascending: bool = False,
    ):
        self.url = url
        self.port = port
        self.embedder = embedder
        self.n_results = n_results
        self.score_threshold = score_threshold
        self.collection_name = collection_name
        self.ascending = ascending
        self.logger = logger or DefaultLogger()

    async def arun(self, query: str, filter: Any | None = None, **kwargs) -> RetrievalResults:
        client = self.get_async_client()

        embed: EmbeddingResults = await self.embedder.arun(query)

        self.logger.info(f"Retrieve from collection {self.collection_name}...")
        retrieved: RetrievalResults = await self._aretrieve(
            client=client,
            collection_name=self.collection_name,
            query_vector=embed.embeddings[0],
            filter=filter,
            n_results=self.n_results,
            score_threshold=self.score_threshold,
            **kwargs,
        )

        retrieved.usage = embed.usage  # override

        return retrieved



import copy
import json
from typing import Callable

from openai._types import NOT_GIVEN, NotGiven

from ...base import (
    BaseConversationLengthAdjuster,
    BaseConversationMemory,
    BaseFilter,
    BaseFunctionCallingModule,
)
from ...llm_wrapper import FunctionCallingWrapperModule
from ...result import FunctionCallingResults, ToolCallResponse, ToolOutput
from ...usage import TokenCounter, Usage
from ..conversation_adjuster.truncate import OldConversationTruncationModule
from ..message import OpenAIMessage
from ..model_config import (
    _OLDER_MODEL_CONFIG,
    MODEL_CONFIG,
    MODEL_POINT,
)
from ..openai_utils import get_async_client, get_client, get_token_limit
from ..tools import OpenAIToolConfig, ToolConfig


class FunctionCallingCoreModule(BaseFunctionCallingModule):
    def __init__(
        self,
        api_key_env_name: str,
        model_name: str,
        tools: list[Callable],
        tool_configs: list[ToolConfig],
        api_type: str = "openai",
        api_version: str | None = None,
        endpoint_env_name: str | None = None,
        deployment_id_env_name: str | None = None,
        organization_id_env_name: str | None = None,
        timeout: int = 30,
        max_retries: int = 2,
        max_tokens: int = 2048,
        seed: int | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven = NOT_GIVEN,
        frequency_penalty: float | NotGiven = NOT_GIVEN,
        presence_penalty: float | NotGiven = NOT_GIVEN,
        temperature: float | NotGiven = NOT_GIVEN,
        user: str | NotGiven = NOT_GIVEN,
        system_instruction: str | None = None,
        conversation_length_adjuster: BaseConversationLengthAdjuster | None = None,
    ) -> None:
        assert api_type in ["openai", "azure"], "api_type must be 'openai' or 'azure'."
        if api_type == "azure":
            assert (
                api_version and endpoint_env_name and deployment_id_env_name
            ), "api_version, endpoint_env_name, and deployment_id_env_name must be specified for Azure API."

        self.api_key_env_name = api_key_env_name
        self.organization_id_env_name = organization_id_env_name
        self.api_type = api_type
        self.api_version = api_version
        self.endpoint_env_name = endpoint_env_name
        self.deployment_id_env_name = deployment_id_env_name
        self.model_name = model_name
        self.timeout = timeout
        self.max_retries = max_retries
        self.top_p = top_p
        self.frequency_penalty = frequency_penalty
        self.presence_penalty = presence_penalty
        self.temperature = temperature
        self.user = user

        self.tools = self._set_runnable_tools_dict(tools)

        _tool_names_from_config = {f.name for f in tool_configs}
        assert (
            len(_tool_names_from_config ^ set(self.tools.keys())) == 0
        ), f"tool names in tool_configs must be the same as the function names in tools. tool names in tool_configs: {_tool_names_from_config}, function names in tools: {set(self.tools.keys())}"

        self.max_tokens = max_tokens

        ClientToolConfig = self._get_client_tool_config_type()
        client_tool_config = ClientToolConfig.from_universal_configs(tool_configs)

        self.additional_inputs = {}
        if model_name not in _OLDER_MODEL_CONFIG.keys():
            self.seed = seed
            self.additional_inputs["seed"] = seed
            self.tool_configs = [f.format() for f in client_tool_config]
            self.additional_inputs["tools"] = self.tool_configs
        else:
            if seed:
                print(
                    f"seed is ignored because it's not supported for {model_name} (api_type:{api_type})"
                )
            self.tool_configs = [f.format()["function"] for f in client_tool_config]
            self.additional_inputs["functions"] = self.tool_configs

        if system_instruction:
            system_instruction = OpenAIMessage.to_universal_message(
                role="system", message=system_instruction
            )
            self.system_instruction = OpenAIMessage.to_client_message(system_instruction)
        else:
            self.system_instruction = None

        self.conversation_length_adjuster = conversation_length_adjuster

    def _get_client_tool_config_type(self) -> ToolConfig:
        return OpenAIToolConfig

    def _set_tool_choice(self, tool_choice: str = "auto"):
        if self.model_name not in _OLDER_MODEL_CONFIG.keys():
            self.additional_inputs["tool_choice"] = (
                str(tool_choice).lower()
                if tool_choice in ["auto", "required", None]
                else {"type": "function", "function": {"name": tool_choice}}
            )
        else:
            self.additional_inputs["function_call"] = tool_choice

    def run(
        self, messages: list[dict[str, str]], tool_choice: str = "auto"
    ) -> FunctionCallingResults:
        self._set_tool_choice(tool_choice)

        if len(messages) == 0:
            raise ValueError("messages must not be empty.")

        client = get_client(
            api_key_env_name=self.api_key_env_name,
            organization_id_env_name=self.organization_id_env_name,
            api_version=self.api_version,
            endpoint_env_name=self.endpoint_env_name,
            deployment_id_env_name=self.deployment_id_env_name,
            api_type=self.api_type,
            max_retries=self.max_retries,
            timeout=self.timeout,
        )

        _messages = [self.system_instruction] + messages if self.system_instruction else messages
        if self.conversation_length_adjuster:
            _messages = self.conversation_length_adjuster.run(_messages)

        response = client.chat.completions.create(
            model=self.model_name,
            messages=_messages,
            temperature=self.temperature,
            max_tokens=self.max_tokens,
            top_p=self.top_p,
            frequency_penalty=self.frequency_penalty,
            presence_penalty=self.presence_penalty,
            stop=NOT_GIVEN,
            stream=False,
            user=self.user,
            **self.additional_inputs,
        )

        usage = Usage(model_name=self.model_name)
        usage += response.usage

        if self.model_name not in _OLDER_MODEL_CONFIG.keys():
            response_message = response.choices[0].message
            self._response_message = response_message
            tool_calls = response_message.tool_calls

            results = []
            calls = []
            if tool_calls is not None:
                for tool_call in tool_calls:
                    call_id = tool_call.id
                    funcname = tool_call.function.name
                    args = tool_call.function.arguments
                    func_out = self.tools[funcname](**json.loads(args))
                    output = ToolOutput(
                        call_id=call_id,
                        funcname=funcname,
                        args=args,
                        output=func_out,
                    )

                    results.append(output)

                    call = ToolCallResponse(
                        call_id=call_id,
                        name=funcname,
                        args=args,
                    )

                    calls.append(call)

            return FunctionCallingResults(
                usage=usage, results=results, prompt=copy.deepcopy(_messages), calls=calls
            )

        elif self.model_name in _OLDER_MODEL_CONFIG.keys():
            response_message = response.choices[0].message
            function_call = response_message.function_call

            output = []
            if function_call is not None:
                funcname = function_call.name
                args = function_call.arguments
                func_out = self.tools[funcname](**json.loads(args))

                output += [
                    ToolOutput(
                        call_id=None,
                        funcname=funcname,
                        args=args,
                        output=func_out,
                    )
                ]

            return FunctionCallingResults(
                usage=usage, results=output, prompt=copy.deepcopy(_messages)
            )

        else:
            raise ValueError(f"model_name {self.model_name} is not supported.")

    async def arun(
        self, messages: list[dict[str, str]], tool_choice: str = "auto"
    ) -> FunctionCallingResults:
        self._set_tool_choice(tool_choice)

        if len(messages) == 0:
            raise ValueError("messages must not be empty.")

        client = get_async_client(
            api_key_env_name=self.api_key_env_name,
            organization_id_env_name=self.organization_id_env_name,
            api_version=self.api_version,
            endpoint_env_name=self.endpoint_env_name,
            deployment_id_env_name=self.deployment_id_env_name,
            api_type=self.api_type,
            max_retries=self.max_retries,
            timeout=self.timeout,
        )

        _messages = [self.system_instruction] + messages if self.system_instruction else messages
        if self.conversation_length_adjuster:
            _messages = self.conversation_length_adjuster.run(_messages)

        response = await client.chat.completions.create(
            model=self.model_name,
            messages=_messages,
            temperature=self.temperature,
            max_tokens=self.max_tokens,
            top_p=self.top_p,
            frequency_penalty=self.frequency_penalty,
            presence_penalty=self.presence_penalty,
            stop=NOT_GIVEN,
            stream=False,
            user=self.user,
            **self.additional_inputs,
        )

        usage = Usage(model_name=self.model_name)
        usage += response.usage

        if self.model_name not in _OLDER_MODEL_CONFIG.keys():
            response_message = response.choices[0].message
            self._response_message = response_message
            tool_calls = response_message.tool_calls

            results = []
            calls = []
            if tool_calls is not None:
                for tool_call in tool_calls:
                    call_id = tool_call.id
                    funcname = tool_call.function.name
                    args = tool_call.function.arguments
                    func_out = self.tools[funcname](**json.loads(args))
                    output = ToolOutput(
                        call_id=call_id,
                        funcname=funcname,
                        args=args,
                        output=func_out,
                    )

                    results.append(output)

                    call = ToolCallResponse(
                        call_id=call_id,
                        name=funcname,
                        args=args,
                    )

                    calls.append(call)

            return FunctionCallingResults(
                usage=usage, results=results, prompt=copy.deepcopy(_messages), calls=calls
            )

        elif self.model_name in _OLDER_MODEL_CONFIG.keys():
            response_message = response.choices[0].message
            function_call = response_message.function_call

            output = []
            if function_call is not None:
                funcname = function_call.name
                args = function_call.arguments
                func_out = self.tools[funcname](**json.loads(args))

                output += [
                    ToolOutput(
                        call_id=None,
                        funcname=funcname,
                        args=args,
                        output=func_out,
                    )
                ]

            return FunctionCallingResults(
                usage=usage, results=output, prompt=copy.deepcopy(_messages)
            )

        else:
            raise ValueError(f"model_name {self.model_name} is not supported.")


class OpenAIFunctionCallingModule(FunctionCallingWrapperModule):
    def __init__(
        self,
        api_key_env_name: str,
        model_name: str,
        tools: list[Callable],
        tool_configs: list[ToolConfig],
        organization_id_env_name: str | None = None,
        api_type: str = "openai",
        api_version: str | None = None,
        endpoint_env_name: str | None = None,
        deployment_id_env_name: str | None = None,
        max_tokens: int = 2048,
        timeout: int = 60,
        max_retries: int = 2,
        seed: int | NotGiven = NOT_GIVEN,
        context_length: int | None = None,
        conversation_memory: BaseConversationMemory | None = None,
        content_filter: BaseFilter | None = None,
        system_instruction: str | None = None,
        conversation_length_adjuster: BaseConversationLengthAdjuster | None = None,
        token_counter: TokenCounter | None = None,
        top_p: float | NotGiven = NOT_GIVEN,
        frequency_penalty: float | NotGiven = NOT_GIVEN,
        presence_penalty: float | NotGiven = NOT_GIVEN,
        temperature: float | NotGiven = NOT_GIVEN,
        user: str | NotGiven = NOT_GIVEN,
    ) -> None:
        if model_name in MODEL_POINT.keys():
            print(f"{model_name} is automatically converted to {MODEL_POINT[model_name]}")
            model_name = MODEL_POINT[model_name]

        assert (
            model_name in MODEL_CONFIG.keys()
        ), f"model_name must be one of {', '.join(sorted(MODEL_CONFIG.keys()))}."

        token_lim = get_token_limit(model_name)
        max_tokens = max_tokens if max_tokens else MODEL_CONFIG[model_name]["max_output_tokens"]
        context_length = token_lim - max_tokens if context_length is None else context_length
        assert (
            token_lim >= max_tokens + context_length
        ), f"max_tokens({max_tokens}) + context_length({context_length}) must be less than or equal to the token limit of the model ({token_lim})."
        assert context_length > 0, "context_length must be positive."

        conversation_length_adjuster = (
            OldConversationTruncationModule(model_name=model_name, context_length=context_length)
            if conversation_length_adjuster is None
            else conversation_length_adjuster
        )

        # The module to call client API
        function_calling_model = FunctionCallingCoreModule(
            api_key_env_name=api_key_env_name,
            organization_id_env_name=organization_id_env_name,
            api_type=api_type,
            api_version=api_version,
            endpoint_env_name=endpoint_env_name,
            deployment_id_env_name=deployment_id_env_name,
            tools=tools,
            tool_configs=tool_configs,
            model_name=model_name,
            max_tokens=max_tokens,
            timeout=timeout,
            max_retries=max_retries,
            seed=seed,
            system_instruction=system_instruction,
            conversation_length_adjuster=conversation_length_adjuster,
            top_p=top_p,
            frequency_penalty=frequency_penalty,
            presence_penalty=presence_penalty,
            temperature=temperature,
            user=user,
        )

        content_filter = content_filter
        conversation_memory = conversation_memory

        super().__init__(
            function_calling_model=function_calling_model,
            conversation_memory=conversation_memory,
            content_filter=content_filter,
            token_counter=token_counter,
        )

    def _get_client_message_type(self) -> OpenAIMessage:
        return OpenAIMessage



from typing import Optional

from pydantic import BaseModel
from vertexai.generative_models import FunctionDeclaration

from ...tools import ToolConfig, ToolParameter, ToolProperty


class VertexAIToolProperty(ToolProperty):
    name: str
    type: str
    description: str
    enum: list[str] | None = None

    def format(self):
        return {self.name: self.model_dump(exclude=["name"], exclude_none=True)}


class VertexAIToolParameter(ToolParameter):
    type: str = "object"
    properties: list[VertexAIToolProperty]
    required: Optional[list[str]] = []

    def format(self):
        properties = {}
        for prop in self.properties:
            prop_dict = prop.format()
            for key, value in prop_dict.items():
                properties.update({key: value})

        return {
            "type": self.type,
            "properties": properties,
            "required": self.required,
        }


class VertexAIToolConfig(ToolConfig):
    name: str
    description: str
    parameters: VertexAIToolParameter

    def format(self):
        return FunctionDeclaration(
            name=self.name,
            description=self.description,
            parameters=self.parameters.format(),
        )

    @classmethod
    def from_universal_configs(cls, configs: list[ToolConfig]) -> list["VertexAIToolConfig"]:
        return [cls(**config.model_dump(exclude_none=True)) for config in configs]



import os
import logging
import json
import hashlib
import secrets
from typing import Optional, Union

import boto3
from botocore.exceptions import ClientError
from botocore.client import Config as BotoConfig

from ..base import BaseConversationMemory

class S3ConversationMemory(BaseConversationMemory):
    def __init__(
            self,
            bucket: str,
            *,
            object_key: Optional[str] = None,
            region_name: Optional[str] = None,
            api_version: Optional[str] = None,
            use_ssl: Optional[bool] = True,
            verify: Union[str, bool, None] = None,
            endpoint_url_env_name: Optional[str] = None,
            aws_access_key_id_env_name: Optional[str] = None,
            aws_secret_access_key_env_name: Optional[str] = None,
            aws_session_token_env_name: Optional[str] = None,
            boto_config: Optional[BotoConfig] = None,
        ):
        self.bucket = bucket
        self.object_key = object_key
        self.region_name = region_name
        self.api_version = api_version
        self.use_ssl = use_ssl
        self.verify = verify
        self.endpoint_url = os.getenv(endpoint_url_env_name) if endpoint_url_env_name else None
        self.aws_access_key_id = os.getenv(aws_access_key_id_env_name) if aws_access_key_id_env_name else None
        self.aws_secret_access_key = os.getenv(aws_secret_access_key_env_name) if aws_secret_access_key_env_name else None
        self.aws_session_token = os.getenv(aws_session_token_env_name) if aws_session_token_env_name else None
        self.boto_config = boto_config

        self.s3_client = boto3.client(
            's3',
            region_name=self.region_name,
            api_version=self.api_version,
            use_ssl=self.use_ssl,
            verify=self.verify,
            endpoint_url=self.endpoint_url,
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
            aws_session_token=self.aws_session_token,
            config=self.boto_config
        )

        try:
            # check at minimum
            self.s3_client.head_bucket(Bucket=self.bucket)
            logging.info(f"s3 connect success. bucket_name: {self.bucket}")
        except Exception as e:
            logging.error(f"s3 connect failed:{e}")
            raise

    def store(self, conversation_history: list[dict[str, str]]):
        if self.object_key is None:
            self.object_key = hashlib.sha256(secrets.token_bytes(32)).hexdigest()
        try:
            json_data = json.dumps(conversation_history, ensure_ascii=False)
            self.s3_client.put_object(Bucket=self.bucket, Key=self.object_key, Body=json_data)
        except Exception as e:
            logging.error(f"s3 store failed: {e}")
            raise

    def load(self) -> list[dict[str, str]]:
        if self.object_key is None:
            return []
        try:
            response = self.s3_client.get_object(Bucket=self.bucket, Key=self.object_key)
            content = response['Body'].read().decode('utf-8')
            conversation_history = json.loads(content)
            return conversation_history
        except ClientError as e:
            error_code = e.response['Error']['Code']
            if error_code == 'NoSuchKey':
                return []
            else:
                logging.error(f"s3 load failed: {e}")
                raise
        except Exception as e:
            logging.error(f"s3 load failedr: {e}")
            raise



from abc import ABC, abstractmethod
from typing import AsyncGenerator, Generator, Optional

from .base import (
    BaseChatModule,
    BaseConversationMemory,
    BaseFilter,
    BaseFunctionCallingModule,
    BaseMessage,
)
from .message_content import ConversationType, InputType, Message
from .mixin import ConversationMixin, FilterMixin
from .result import CompletionResults, FunctionCallingResults
from .usage import TokenCounter


class ChatWrapperModule(ABC, ConversationMixin, FilterMixin):
    def __init__(
        self,
        chat_model: BaseChatModule,
        conversation_memory: BaseConversationMemory | None = None,
        content_filter: BaseFilter | None = None,
        token_counter: TokenCounter | None = None,
    ):
        self.chat_model = chat_model
        self.conversation_memory = conversation_memory
        self.content_filter = content_filter
        self.token_counter = token_counter
        self._INIT_STATUS = False

    @abstractmethod
    def _get_client_message_type(self) -> type[BaseMessage]:
        raise NotImplementedError

    def run(
        self,
        prompt: InputType,
        init_conversation: ConversationType | None = None,
        **kwargs,
    ) -> CompletionResults:
        ClientMessage = self._get_client_message_type()

        # Load conversation memory
        messages = self.load_conversation()

        # Convert init_conversation to ClientMessage
        if init_conversation:
            if not isinstance(init_conversation, list):
                init_conversation = [init_conversation]

            messages.extend(init_conversation)

        # Convert messages to UniversalMessage
        messages = [ClientMessage.to_universal_message(message=message) for message in messages]

        # Convert prompt to UniversalMessage
        messages.append(ClientMessage.to_universal_message(message=prompt, role="user"))

        # Convert UniversalMessage to ClientMessage
        _messages = [
            ClientMessage.to_client_message(message=message)
            for message in ClientMessage._preprocess_message(messages)
        ]

        # Apply content filter if available
        if self.content_filter is not None:
            _messages = self.content_filter.apply(_messages)

        response: CompletionResults = self.chat_model.run(_messages, **kwargs)

        # Update total tokens
        if self.token_counter is not None:
            self.token_counter += response.usage

        # Restore content filter if available
        if self.content_filter is not None:
            response.message = self.restore_content_filter([response.message])[0]

        # Append the response to the conversation with the UniversalMessage format
        response.message = ClientMessage.to_universal_message_from_completion_response(
            response=response
        )
        messages.append(response.message)

        # Save conversation memory
        if self.conversation_memory is not None:
            serializable = [m.model_dump() for m in messages]
            self.save_conversation(serializable)

        return response

    async def arun(
        self,
        prompt: InputType,
        init_conversation: ConversationType | None = None,
        **kwargs,
    ) -> CompletionResults:
        ClientMessage = self._get_client_message_type()

        # Load conversation memory
        messages = self.load_conversation()

        # Convert init_conversation to ClientMessage
        if init_conversation:
            if not isinstance(init_conversation, list):
                init_conversation = [init_conversation]

            messages.extend(init_conversation)

        # Convert messages to UniversalMessage
        messages = [ClientMessage.to_universal_message(message=message) for message in messages]

        # Convert prompt to UniversalMessage
        messages.append(ClientMessage.to_universal_message(message=prompt, role="user"))

        # Convert UniversalMessage to ClientMessage
        _messages = [
            ClientMessage.to_client_message(message=message)
            for message in ClientMessage._preprocess_message(messages)
        ]

        # Apply content filter if available
        if self.content_filter is not None:
            _messages = self.content_filter.apply(_messages)

        response: CompletionResults = await self.chat_model.arun(_messages, **kwargs)

        # Update total tokens
        if self.token_counter is not None:
            self.token_counter += response.usage

        # Restore content filter if available
        if self.content_filter is not None:
            response.message = self.restore_content_filter([response.message])[0]

        # Append the response to the conversation with the UniversalMessage format
        response.message = ClientMessage.to_universal_message_from_completion_response(
            response=response
        )
        messages.append(response.message)

        # Save conversation memory
        if self.conversation_memory is not None:
            serializable = [m.model_dump() for m in messages]
            self.save_conversation(serializable)

        return response

    def stream(
        self,
        prompt: InputType,
        init_conversation: ConversationType | None = None,
    ) -> Generator[CompletionResults, None, None]:
        ClientMessage = self._get_client_message_type()

        # Load conversation memory
        messages = self.load_conversation()

        # Convert init_conversation to ClientMessage
        if init_conversation:
            if not isinstance(init_conversation, list):
                init_conversation = [init_conversation]

            messages.extend(init_conversation)

        # Convert messages to UniversalMessage
        messages = [ClientMessage.to_universal_message(message=message) for message in messages]

        # Convert prompt to UniversalMessage
        messages.append(ClientMessage.to_universal_message(message=prompt, role="user"))

        # Convert UniversalMessage to ClientMessage
        _messages = [
            ClientMessage.to_client_message(message=message)
            for message in ClientMessage._preprocess_message(messages)
        ]

        # Apply content filter if available
        if self.content_filter is not None:
            _messages = self.content_filter.apply(_messages)

        # Run the model
        response = self.chat_model.stream(_messages)

        for chunk in response:
            if isinstance(chunk, CompletionResults):
                if self.content_filter is not None:
                    chunk.message = self.restore_content_filter([chunk.message])[0]

                # Convert to UniversalMessage
                chunk.message = ClientMessage.to_universal_message_from_completion_response(
                    response=chunk
                )

                yield chunk
            else:
                raise AssertionError

        # Update total tokens
        if self.token_counter is not None:
            self.token_counter += chunk.usage

        messages.append(chunk.message)

        # Save conversation memory
        if self.conversation_memory is not None:
            serializable = [m.model_dump() for m in messages]
            self.save_conversation(serializable)

    async def astream(
        self,
        prompt: InputType,
        init_conversation: ConversationType | None = None,
    ) -> AsyncGenerator[CompletionResults, None]:
        ClientMessage = self._get_client_message_type()

        # Load conversation memory
        messages = self.load_conversation()

        # Convert init_conversation to ClientMessage
        if init_conversation:
            if not isinstance(init_conversation, list):
                init_conversation = [init_conversation]

            messages.extend(init_conversation)

        # Convert messages to UniversalMessage
        messages = [ClientMessage.to_universal_message(message=message) for message in messages]

        # Convert prompt to UniversalMessage
        messages.append(ClientMessage.to_universal_message(message=prompt, role="user"))

        # Convert UniversalMessage to ClientMessage
        _messages = [
            ClientMessage.to_client_message(message=message)
            for message in ClientMessage._preprocess_message(messages)
        ]

        # Apply content filter if available
        if self.content_filter is not None:
            _messages = self.content_filter.apply(_messages)

        # Run the model
        response = self.chat_model.astream(_messages)

        async for chunk in response:
            if isinstance(chunk, CompletionResults):
                if self.content_filter is not None:
                    chunk.message = self.restore_content_filter([chunk.message])[0]

                # Convert to UniversalMessage
                chunk.message = ClientMessage.to_universal_message_from_completion_response(
                    response=chunk
                )

                yield chunk
            else:
                raise AssertionError

        # Update total tokens
        if self.token_counter is not None:
            self.token_counter += chunk.usage

        messages.append(chunk.message)

        # Save conversation memory
        if self.conversation_memory is not None:
            serializable = [m.model_dump() for m in messages]
            self.save_conversation(serializable)


class FunctionCallingWrapperModule(ABC, ConversationMixin, FilterMixin):
    def __init__(
        self,
        function_calling_model: BaseFunctionCallingModule,
        conversation_memory: Optional[BaseConversationMemory] = None,
        content_filter: Optional[BaseFilter] = None,
        token_counter: Optional[TokenCounter] = None,
    ):
        self.function_calling_model = function_calling_model
        self.conversation_memory = conversation_memory
        self.content_filter = content_filter
        self.token_counter = token_counter
        self._INIT_STATUS = False

    @abstractmethod
    def _get_client_message_type(self) -> type[BaseMessage]:
        raise NotImplementedError

    def run(
        self,
        prompt: InputType,
        init_conversation: ConversationType | None = None,
        **kwargs,
    ) -> FunctionCallingResults:
        ClientMessage = self._get_client_message_type()

        # Load conversation memory
        messages = self.load_conversation()

        # Convert init_conversation to ClientMessage
        if init_conversation:
            if not isinstance(init_conversation, list):
                init_conversation = [init_conversation]

            messages.extend(init_conversation)

        # Convert messages to UniversalMessage
        messages = [ClientMessage.to_universal_message(message=message) for message in messages]

        # Convert prompt to UniversalMessage
        messages.append(ClientMessage.to_universal_message(message=prompt, role="user"))

        # Convert UniversalMessage to ClientMessage
        _messages = [
            ClientMessage.to_client_message(message=message)
            for message in ClientMessage._preprocess_message(messages)
        ]

        # Apply content filter if available
        if self.content_filter is not None:
            _messages = self.content_filter.apply(_messages)

        # Run the model
        response = self.function_calling_model.run(_messages, **kwargs)

        # Update total tokens
        if self.token_counter is not None:
            self.token_counter += response.usage

        # Restore content filter if available
        if self.content_filter is not None:
            for i, _ in enumerate(response.results):
                response.results[i].args = self.restore_content_filter([response.results[i].args])[
                    0
                ]

        # Append the response to the conversation
        response.results = ClientMessage.to_universal_message_from_function_response(
            response=response
        )

        if response.calls:
            response.calls = ClientMessage.to_universal_message_from_function_call(
                response=response
            )
            messages.append(response.calls)
        else:
            messages.extend(response.results)

        # Save conversation memory
        if self.conversation_memory is not None and (response.calls or response.results):
            serializable = [m.model_dump() for m in messages]
            self.save_conversation(serializable)

        return response

    async def arun(
        self,
        prompt: InputType,
        init_conversation: ConversationType | None = None,
        **kwargs,
    ) -> FunctionCallingResults:
        ClientMessage = self._get_client_message_type()

        # Load conversation memory
        messages = self.load_conversation()

        # Convert init_conversation to ClientMessage
        if init_conversation:
            if not isinstance(init_conversation, list):
                init_conversation = [init_conversation]

            messages.extend(init_conversation)

        # Convert messages to UniversalMessage
        messages = [ClientMessage.to_universal_message(message=message) for message in messages]

        # Convert prompt to UniversalMessage
        messages.append(ClientMessage.to_universal_message(message=prompt, role="user"))

        # Convert UniversalMessage to ClientMessage
        _messages = [
            ClientMessage.to_client_message(message=message)
            for message in ClientMessage._preprocess_message(messages)
        ]

        # Apply content filter if available
        if self.content_filter is not None:
            _messages = self.content_filter.apply(_messages)

        # Run the model
        response = await self.function_calling_model.arun(_messages, **kwargs)

        # Update total tokens
        if self.token_counter is not None:
            self.token_counter += response.usage

        # Restore content filter if available
        if self.content_filter is not None:
            for i, _ in enumerate(response.results):
                response.results[i].args = self.restore_content_filter([response.results[i].args])[
                    0
                ]

        # Append the response to the conversation
        response.results = ClientMessage.to_universal_message_from_function_response(
            response=response
        )

        if response.calls:
            response.calls = ClientMessage.to_universal_message_from_function_call(
                response=response
            )
            messages.append(response.calls)
        else:
            messages.extend(response.results)

        # Save conversation memory
        if self.conversation_memory is not None and (response.calls or response.results):
            serializable = [m.model_dump() for m in messages]
            self.save_conversation(serializable)

        return response



from abc import ABC, abstractmethod
from typing import AsyncGenerator, Generator, Literal

from .base import BaseConversationMemory
from .memory.in_memory import InMemoryConversationMemory
from .message_content import ConversationType, InputType
from .result import CompletionResults


class BaseAssembly(ABC):
    def __init__(self, conversation_memory: BaseConversationMemory | None = None):
        if conversation_memory:
            self.conversation_memory = conversation_memory
            self.__HASMEMORY = True
        else:
            # conversation memory is useful to bridge chat and function calling
            # so if it is not provided, we will use InMemoryConversationMemory internally,
            # and clear the memory after each run
            self.conversation_memory = InMemoryConversationMemory()
            self.__HASMEMORY = False

    @abstractmethod
    def run(
        self,
        prompt: InputType,
        init_conversation: ConversationType | None = None,
        tool_choice: Literal["auto", "required", "any"] | str | None = None,
        tool_only: bool = False,
    ) -> CompletionResults:
        raise NotImplementedError

    async def arun(
        self,
        prompt: InputType,
        init_conversation: ConversationType | None = None,
        tool_choice: Literal["auto", "required", "any"] | str | None = None,
        tool_only: bool = False,
    ) -> CompletionResults:
        raise NotImplementedError

    def stream(
        self,
        prompt: InputType,
        init_conversation: ConversationType | None = None,
        tool_choice: Literal["auto", "required", "any"] | str | None = None,
    ) -> Generator[CompletionResults, None, None]:
        raise NotImplementedError

    async def astream(
        self,
        prompt: InputType,
        init_conversation: ConversationType | None = None,
        tool_choice: Literal["auto", "required", "any"] | str | None = None,
    ) -> AsyncGenerator[CompletionResults, None]:
        raise NotImplementedError

    def _clear_memory(self) -> None:
        if not self.__HASMEMORY:
            # Clear conversation memory if conversation memory is not provided
            self.conversation_memory.store([])



from typing import AsyncGenerator, Callable, Generator, Literal

from openai._types import NOT_GIVEN, NotGiven
from pydantic import BaseModel

from ..base import BaseConversationLengthAdjuster, BaseConversationMemory, BaseFilter
from ..base_assembly import BaseAssembly
from ..message_content import ConversationType, InputType, Message, ToolContent
from ..result import CompletionResults, FunctionCallingResults
from ..tools import ToolConfig
from ..usage import TokenCounter, Usage
from .llm.chat import OpenAIChatModule
from .llm.function_calling import OpenAIFunctionCallingModule


class OpenAIFunctionalChat(BaseAssembly):
    def __init__(
        self,
        api_key_env_name: str,
        model_name: str,
        tools: list[Callable] | None = None,
        tool_configs: list[ToolConfig] | None = None,
        organization_id_env_name: str | None = None,
        api_type: str = "openai",
        api_version: str | None = None,
        endpoint_env_name: str | None = None,
        deployment_id_env_name: str | None = None,
        max_tokens: int = 2048,
        timeout: int = 60,
        max_retries: int = 2,
        seed: int | NotGiven = NOT_GIVEN,
        context_length: int | None = None,
        conversation_memory: BaseConversationMemory | None = None,
        content_filter: BaseFilter | None = None,
        system_instruction: str | None = None,
        conversation_length_adjuster: BaseConversationLengthAdjuster | None = None,
        token_counter: TokenCounter | None = None,
        json_mode: bool = False,
        top_p: float | None | NotGiven = NOT_GIVEN,
        frequency_penalty: float | NotGiven = NOT_GIVEN,
        presence_penalty: float | None | NotGiven = NOT_GIVEN,
        temperature: float | None | NotGiven = NOT_GIVEN,
        user: str | NotGiven = NOT_GIVEN,
        response_schema: BaseModel | None = None,
    ) -> None:
        super().__init__(conversation_memory=conversation_memory)

        self.chat = OpenAIChatModule(
            api_key_env_name=api_key_env_name,
            model_name=model_name,
            organization_id_env_name=organization_id_env_name,
            api_type=api_type,
            api_version=api_version,
            endpoint_env_name=endpoint_env_name,
            deployment_id_env_name=deployment_id_env_name,
            max_tokens=max_tokens,
            timeout=timeout,
            max_retries=max_retries,
            seed=seed,
            context_length=context_length,
            conversation_memory=self.conversation_memory,
            content_filter=content_filter,
            system_instruction=system_instruction,
            conversation_length_adjuster=conversation_length_adjuster,
            token_counter=token_counter,
            json_mode=json_mode,
            top_p=top_p,
            frequency_penalty=frequency_penalty,
            presence_penalty=presence_penalty,
            temperature=temperature,
            user=user,
            response_schema=response_schema,
        )

        if tools:
            self.function_calling = OpenAIFunctionCallingModule(
                api_key_env_name=api_key_env_name,
                model_name=model_name,
                tools=tools,
                tool_configs=tool_configs,
                organization_id_env_name=organization_id_env_name,
                api_type=api_type,
                api_version=api_version,
                endpoint_env_name=endpoint_env_name,
                deployment_id_env_name=deployment_id_env_name,
                max_tokens=max_tokens,
                timeout=timeout,
                max_retries=max_retries,
                seed=seed,
                context_length=context_length,
                conversation_memory=self.conversation_memory,
                content_filter=content_filter,
                system_instruction=system_instruction,
                conversation_length_adjuster=conversation_length_adjuster,
                token_counter=token_counter,
                top_p=top_p,
                frequency_penalty=frequency_penalty,
                presence_penalty=presence_penalty,
                temperature=temperature,
                user=user,
            )
        else:
            self.function_calling = None

    def run(
        self,
        prompt: InputType,
        init_conversation: ConversationType | None = None,
        tool_choice: Literal["auto", "required"] | str | None = None,
        tool_only: bool = False,
        n_results: int | NotGiven = NOT_GIVEN,
    ) -> CompletionResults | FunctionCallingResults:
        if self.function_calling and tool_choice is None:
            tool_choice = "auto"

        if tool_only:
            assert tool_choice is not None, "tool_choice must be provided when tool_only is True"

        total_usage = Usage(model_name=self.chat.chat_model.model_name)

        if self.function_calling and tool_choice:
            response_function_calling: FunctionCallingResults = self.function_calling.run(
                prompt=prompt,
                init_conversation=init_conversation,
                tool_choice=tool_choice,
            )

            if tool_only:
                self._clear_memory()

                return response_function_calling

            if response_function_calling.results:
                prompt = Message(
                    role="function",
                    name=None,
                    content=[
                        ToolContent(**content.model_dump())
                        for result in response_function_calling.results
                        for content in result.content
                    ],
                )
                init_conversation = (
                    None  # if tool is used, init_conversation is stored in the memory
                )

            total_usage += response_function_calling.usage

        response_chat: CompletionResults = self.chat.run(
            prompt=prompt,
            n_results=n_results,
            init_conversation=init_conversation,
        )

        total_usage += response_chat.usage
        response_chat.usage = total_usage

        self._clear_memory()

        return response_chat

    async def arun(
        self,
        prompt: InputType,
        init_conversation: ConversationType | None = None,
        tool_choice: Literal["auto", "required"] | str | None = None,
        tool_only: bool = False,
        n_results: int | NotGiven = NOT_GIVEN,
    ) -> CompletionResults | FunctionCallingResults:
        if self.function_calling and tool_choice is None:
            tool_choice = "auto"

        if tool_only:
            assert tool_choice is not None, "tool_choice must be provided when tool_only is True"

        total_usage = Usage(model_name=self.chat.chat_model.model_name)

        if self.function_calling and tool_choice:
            response_function_calling: FunctionCallingResults = await self.function_calling.arun(
                prompt=prompt,
                init_conversation=init_conversation,
                tool_choice=tool_choice,
            )

            if tool_only:
                self._clear_memory()

                return response_function_calling

            if response_function_calling.results:
                prompt = Message(
                    role="function",
                    name=None,
                    content=[
                        ToolContent(**content.model_dump())
                        for result in response_function_calling.results
                        for content in result.content
                    ],
                )
                init_conversation = (
                    None  # if tool is used, init_conversation is stored in the memory
                )

            total_usage += response_function_calling.usage

        response_chat: CompletionResults = await self.chat.arun(
            prompt=prompt,
            n_results=n_results,
            init_conversation=init_conversation,
        )

        total_usage += response_chat.usage
        response_chat.usage = total_usage

        self._clear_memory()

        return response_chat

    def stream(
        self,
        prompt: InputType,
        init_conversation: ConversationType | None = None,
        tool_choice: Literal["auto", "required"] | str | None = None,
    ) -> Generator[CompletionResults, None, None]:
        if self.function_calling and tool_choice is None:
            tool_choice = "auto"

        total_usage = Usage(model_name=self.chat.chat_model.model_name)

        if self.function_calling and tool_choice:
            response_function_calling: FunctionCallingResults = self.function_calling.run(
                prompt=prompt,
                init_conversation=init_conversation,
                tool_choice=tool_choice,
            )

            if response_function_calling.results:
                prompt = Message(
                    role="function",
                    name=None,
                    content=[
                        ToolContent(**content.model_dump())
                        for result in response_function_calling.results
                        for content in result.content
                    ],
                )
                init_conversation = (
                    None  # if tool is used, init_conversation is stored in the memory
                )

            total_usage += response_function_calling.usage

        response_chat = self.chat.stream(
            prompt=prompt,
            init_conversation=init_conversation,
        )

        for result in response_chat:
            if result.usage.total_tokens > 0:
                total_usage += result.usage
                result.usage = total_usage
            yield result

        self._clear_memory()

    async def astream(
        self,
        prompt: InputType,
        init_conversation: ConversationType | None = None,
        tool_choice: Literal["auto", "required"] | str | None = None,
    ) -> AsyncGenerator[CompletionResults, None]:
        if self.function_calling and tool_choice is None:
            tool_choice = "auto"

        total_usage = Usage(model_name=self.chat.chat_model.model_name)

        if self.function_calling and tool_choice:
            response_function_calling: FunctionCallingResults = await self.function_calling.arun(
                prompt=prompt,
                init_conversation=init_conversation,
                tool_choice=tool_choice,
            )

            if response_function_calling.results:
                prompt = Message(
                    role="function",
                    name=None,
                    content=[
                        ToolContent(**content.model_dump())
                        for result in response_function_calling.results
                        for content in result.content
                    ],
                )
                init_conversation = (
                    None  # if tool is used, init_conversation is stored in the memory
                )

            total_usage += response_function_calling.usage

        response_chat = self.chat.astream(
            prompt=prompt,
            init_conversation=init_conversation,
        )

        async for result in response_chat:
            if result.usage.total_tokens > 0:
                total_usage += result.usage
                result.usage = total_usage
            yield result

        self._clear_memory()



from typing import Optional

from google.ai.generativelanguage import FunctionDeclaration, Schema, Type

from ...tools import ToolConfig, ToolParameter, ToolProperty


class GeminiToolProperty(ToolProperty):
    name: str
    type: str
    description: str
    enum: list[str] | None = None

    def format(self):
        return {
            self.name: Schema(
                type=getattr(Type, self.type.upper()),
                **self.model_dump(exclude=["name", "type"], exclude_none=True),
            ),
        }


class GeminiToolParameter(ToolParameter):
    type: str = "object"
    properties: list[GeminiToolProperty]
    required: Optional[list[str]] = []

    def format(self):
        properties = {}
        for prop in self.properties:
            prop_dict = prop.format()
            for key, value in prop_dict.items():
                properties.update({key: value})

        return Schema(
            type=getattr(Type, self.type.upper()),
            properties=properties,
            required=self.required,
        )


class GeminiToolConfig(ToolConfig):
    name: str
    description: str
    parameters: GeminiToolParameter

    def format(self):
        return FunctionDeclaration(
            name=self.name,
            description=self.description,
            parameters=self.parameters.format(),
        )

    @classmethod
    def from_universal_configs(cls, configs: list[ToolConfig]) -> list["GeminiToolConfig"]:
        return [cls(**config.model_dump(exclude_none=True)) for config in configs]



import os
from typing import Any, Callable, Optional, Sequence

import vertexai
from google.auth import credentials as auth_credentials
from google.cloud.aiplatform.tensorboard import tensorboard_resource
from vertexai.generative_models import GenerationConfig, GenerativeModel


def get_vertexai_model(
    model_name: str,
    project_id_env_name: str,
    location_env_name: str,
    n_results: int = 1,
    system_instruction: str | None = None,
    max_output_tokens: int = 2048,
    json_mode: bool = False,
    experiment: str | None = None,
    experiment_description: str | None = None,
    experiment_tensorboard: str | tensorboard_resource.Tensorboard | bool | None = None,
    staging_bucket: str | None = None,
    credentials: auth_credentials.Credentials | None = None,
    encryption_spec_key_name: str | None = None,
    network: str | None = None,
    service_account: str | None = None,
    endpoint_env_name: str | None = None,
    request_metadata: Sequence[tuple[str, str]] | None = None,
    response_schema: dict[str, Any] | None = None,
    presence_penalty: float | None = None,
    frequency_penalty: float | None = None,
    temperature: float | None = None,
    top_p: float | None = None,
    top_k: int | None = None,
):
    vertexai.init(
        project=os.getenv(project_id_env_name),
        location=os.getenv(location_env_name),
        experiment=experiment,
        experiment_description=experiment_description,
        experiment_tensorboard=experiment_tensorboard,
        staging_bucket=staging_bucket,
        credentials=credentials,
        encryption_spec_key_name=encryption_spec_key_name,
        network=network,
        service_account=service_account,
        api_endpoint=os.getenv(endpoint_env_name) if endpoint_env_name else endpoint_env_name,
        request_metadata=request_metadata,
    )

    generation_config = GenerationConfig(
        candidate_count=n_results,
        stop_sequences=None,
        max_output_tokens=max_output_tokens,
        temperature=temperature,
        top_p=top_p,
        top_k=top_k,
        response_mime_type="text/plain" if not json_mode else "application/json",
        response_schema=response_schema,
        presence_penalty=presence_penalty,
        frequency_penalty=frequency_penalty,
    )

    return GenerativeModel(
        model_name=model_name,
        system_instruction=system_instruction,
        generation_config=generation_config,
    )



from typing import Any, AsyncGenerator, Callable, Generator, Literal, Sequence

from google.auth import credentials as auth_credentials

from ..base import (
    BaseConversationMemory,
    BaseFilter,
)
from ..base_assembly import BaseAssembly
from ..message_content import ConversationType, InputType, Message, ToolContent
from ..result import CompletionResults, FunctionCallingResults
from ..tools import ToolConfig
from ..usage import TokenCounter, Usage
from .llm.chat import GeminiChatModule
from .llm.function_calling import GeminiFunctionCallingModule


class GeminiFunctionalChat(BaseAssembly):
    def __init__(
        self,
        model_name: str,
        api_key_env_name: str | None = None,
        max_tokens: int = 2048,
        json_mode: bool = False,
        timeout: int = 60,
        content_filter: BaseFilter | None = None,
        conversation_memory: BaseConversationMemory | None = None,
        system_instruction: str | None = None,
        token_counter: TokenCounter | None = None,
        api_type: str = "genai",
        project_id_env_name: str | None = None,
        location_env_name: str | None = None,
        experiment: str | None = None,
        experiment_description: str | None = None,
        experiment_tensorboard: str | bool | None = None,
        staging_bucket: str | None = None,
        credentials: auth_credentials.Credentials | None = None,
        encryption_spec_key_name: str | None = None,
        network: str | None = None,
        service_account: str | None = None,
        endpoint_env_name: str | None = None,
        request_metadata: Sequence[tuple[str, str]] | None = None,
        tools: list[Callable] | None = None,
        tool_configs: list[ToolConfig] | None = None,
        response_schema: dict[str, Any] | None = None,
        presence_penalty: float | None = None,
        frequency_penalty: float | None = None,
        temperature: float | None = None,
        top_p: float | None = None,
        top_k: int | None = None,
    ):
        super().__init__(conversation_memory=conversation_memory)

        self.chat = GeminiChatModule(
            model_name=model_name,
            api_key_env_name=api_key_env_name,
            max_tokens=max_tokens,
            json_mode=json_mode,
            timeout=timeout,
            content_filter=content_filter,
            conversation_memory=self.conversation_memory,
            system_instruction=system_instruction,
            token_counter=token_counter,
            api_type=api_type,
            project_id_env_name=project_id_env_name,
            location_env_name=location_env_name,
            experiment=experiment,
            experiment_description=experiment_description,
            experiment_tensorboard=experiment_tensorboard,
            staging_bucket=staging_bucket,
            credentials=credentials,
            encryption_spec_key_name=encryption_spec_key_name,
            network=network,
            service_account=service_account,
            endpoint_env_name=endpoint_env_name,
            request_metadata=request_metadata,
            response_schema=response_schema,
            presence_penalty=presence_penalty,
            frequency_penalty=frequency_penalty,
            temperature=temperature,
            top_p=top_p,
            top_k=top_k,
        )

        if tools:
            self.function_calling = GeminiFunctionCallingModule(
                model_name=model_name,
                api_key_env_name=api_key_env_name,
                max_tokens=max_tokens,
                json_mode=json_mode,
                timeout=timeout,
                content_filter=content_filter,
                conversation_memory=self.conversation_memory,
                system_instruction=system_instruction,
                token_counter=token_counter,
                api_type=api_type,
                project_id_env_name=project_id_env_name,
                location_env_name=location_env_name,
                experiment=experiment,
                experiment_description=experiment_description,
                experiment_tensorboard=experiment_tensorboard,
                staging_bucket=staging_bucket,
                credentials=credentials,
                encryption_spec_key_name=encryption_spec_key_name,
                network=network,
                service_account=service_account,
                endpoint_env_name=endpoint_env_name,
                request_metadata=request_metadata,
                tools=tools,
                tool_configs=tool_configs,
                presence_penalty=presence_penalty,
                frequency_penalty=frequency_penalty,
                temperature=temperature,
                top_p=top_p,
                top_k=top_k,
            )
        else:
            self.function_calling = None

    def run(
        self,
        prompt: InputType,
        init_conversation: ConversationType | None = None,
        tool_choice: Literal["auto", "any"] | str | None = None,
        tool_only: bool = False,
        n_results: int | None = None,
    ) -> CompletionResults | FunctionCallingResults:
        if self.function_calling and tool_choice is None:
            tool_choice = "auto"

        if tool_only:
            assert tool_choice is not None, "tool_choice must be provided when tool_only is True"

        total_usage = Usage(model_name=self.chat.chat_model.model_name)

        if self.function_calling and tool_choice:
            response_function_calling: FunctionCallingResults = self.function_calling.run(
                prompt,
                init_conversation=init_conversation,
                tool_choice=tool_choice,
            )

            if tool_only:
                self._clear_memory()

                return response_function_calling

            if response_function_calling.results:
                prompt = Message(
                    role="function",
                    content=[
                        ToolContent(**content.model_dump())
                        for result in response_function_calling.results
                        for content in result.content
                    ],
                )
                init_conversation = (
                    None  # if tool is used, init_conversation is stored in the memory
                )

            total_usage += response_function_calling.usage

        response_chat: CompletionResults = self.chat.run(
            prompt, init_conversation=init_conversation, n_results=n_results
        )

        total_usage += response_chat.usage
        response_chat.usage = total_usage

        self._clear_memory()

        return response_chat

    async def arun(
        self,
        prompt: InputType,
        init_conversation: ConversationType | None = None,
        tool_choice: Literal["auto", "any"] | str | None = None,
        tool_only: bool = False,
        n_results: int | None = None,
    ) -> CompletionResults | FunctionCallingResults:
        if self.function_calling and tool_choice is None:
            tool_choice = "auto"

        if tool_only:
            assert tool_choice is not None, "tool_choice must be provided when tool_only is True"

        total_usage = Usage(model_name=self.chat.chat_model.model_name)

        if self.function_calling and tool_choice:
            response_function_calling: FunctionCallingResults = await self.function_calling.arun(
                prompt,
                init_conversation=init_conversation,
                tool_choice=tool_choice,
            )

            if tool_only:
                self._clear_memory()

                return response_function_calling

            if response_function_calling.results:
                prompt = Message(
                    role="function",
                    content=[
                        ToolContent(**content.model_dump())
                        for result in response_function_calling.results
                        for content in result.content
                    ],
                )
                init_conversation = (
                    None  # if tool is used, init_conversation is stored in the memory
                )

            total_usage += response_function_calling.usage

        response_chat: CompletionResults = await self.chat.arun(
            prompt, init_conversation=init_conversation, n_results=n_results
        )

        total_usage += response_chat.usage
        response_chat.usage = total_usage

        self._clear_memory()

        return response_chat

    def stream(
        self,
        prompt: InputType,
        init_conversation: ConversationType | None = None,
        tool_choice: Literal["auto", "any"] | str | None = None,
    ) -> Generator[CompletionResults, None, None]:
        if self.function_calling and tool_choice is None:
            tool_choice = "auto"

        total_usage = Usage(model_name=self.chat.chat_model.model_name)

        if self.function_calling and tool_choice:
            response_function_calling: FunctionCallingResults = self.function_calling.run(
                prompt,
                init_conversation=init_conversation,
                tool_choice=tool_choice,
            )

            if response_function_calling.results:
                prompt = Message(
                    role="function",
                    content=[
                        ToolContent(**content.model_dump())
                        for result in response_function_calling.results
                        for content in result.content
                    ],
                )
                init_conversation = (
                    None  # if tool is used, init_conversation is stored in the memory
                )

            total_usage += response_function_calling.usage

        response_chat = self.chat.stream(prompt, init_conversation=init_conversation)

        for result in response_chat:
            if result.usage.total_tokens > 0:
                total_usage += result.usage
                result.usage = total_usage
            yield result

        self._clear_memory()

    async def astream(
        self,
        prompt: InputType,
        init_conversation: ConversationType | None = None,
        tool_choice: Literal["auto", "any"] | str | None = None,
    ) -> AsyncGenerator[CompletionResults, None]:
        if self.function_calling and tool_choice is None:
            tool_choice = "auto"

        total_usage = Usage(model_name=self.chat.chat_model.model_name)

        if self.function_calling and tool_choice:
            response_function_calling: FunctionCallingResults = await self.function_calling.arun(
                prompt,
                init_conversation=init_conversation,
                tool_choice=tool_choice,
            )

            if response_function_calling.results:
                prompt = Message(
                    role="function",
                    content=[
                        ToolContent(**content.model_dump())
                        for result in response_function_calling.results
                        for content in result.content
                    ],
                )
                init_conversation = (
                    None  # if tool is used, init_conversation is stored in the memory
                )

            total_usage += response_function_calling.usage

        response_chat = self.chat.astream(prompt, init_conversation=init_conversation)

        async for result in response_chat:
            if result.usage.total_tokens > 0:
                total_usage += result.usage
                result.usage = total_usage
            yield result

        self._clear_memory()



MODEL_CONFIG = {
    "models/gemini-1.5-pro": {
        "context_tokens": 1048576,
        "max_tokens": 8192,
        "prompt_cost_per_token": 0.0000035,
        "completion_cost_per_token": 0.0000105,
    },
    "models/gemini-1.5-flash": {
        "context_tokens": 1048576,
        "max_tokens": 8192,
        "prompt_cost_per_token": 0.00000035,
        "completion_cost_per_token": 0.00000105,
    },
}



import copy
import json
from typing import Any, Callable, Iterable, Mapping, Union

import httpx
from anthropic._base_client import DEFAULT_MAX_RETRIES
from anthropic._types import (
    NOT_GIVEN,
    NotGiven,
    ProxiesTypes,
    Timeout,
    Transport,
)
from anthropic.types import TextBlockParam, ToolUseBlock
from anthropic.types.message_create_params import (
    ToolChoice,
    ToolChoiceToolChoiceAny,
    ToolChoiceToolChoiceAuto,
    ToolChoiceToolChoiceTool,
)
from anthropic.types.text_block import TextBlock

from ...base import (
    BaseChatModule,
    BaseConversationLengthAdjuster,
    BaseConversationMemory,
    BaseFilter,
    BaseFunctionCallingModule,
    BaseMessage,
)
from ...llm_wrapper import FunctionCallingWrapperModule
from ...message_content import TextContent
from ...result import FunctionCallingResults, ToolCallResponse, ToolOutput
from ...tools import ToolConfig
from ...usage import TokenCounter, Usage
from ..claude_utils import acompletion, completion, get_async_client, get_client
from ..message import ClaudeMessage
from ..tools import ClaudeToolConfig


class AnthropicFunctionCallingCoreModule(BaseFunctionCallingModule):
    def __init__(
        self,
        model_name: str,
        tools: list[Callable],
        tool_configs: list[ToolConfig],
        api_type: str = "anthropic",
        api_key_env_name: str | None = None,
        auth_token_env_name: str | None = None,
        endpoint_env_name: str | httpx.URL | None = None,
        aws_secret_key_env_name: str | None = None,
        aws_access_key_env_name: str | None = None,
        aws_region_env_name: str | None = None,
        aws_session_token_env_name: str | None = None,
        gc_region_env_name: str | NotGiven = NOT_GIVEN,
        gc_project_id_env_name: str | NotGiven = NOT_GIVEN,
        gc_access_token_env_name: str | None = None,
        credentials: Any | None = None,
        timeout: Union[float, Timeout, None, NotGiven] = 600,
        max_tokens: int = 2048,
        max_retries: int = DEFAULT_MAX_RETRIES,
        default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        http_client: httpx.Client | None = None,
        transport: Transport | None = None,
        proxies: ProxiesTypes | None = None,
        connection_pool_limits: httpx.Limits | None = None,
        _strict_response_validation: bool = False,
        conversation_length_adjuster: BaseConversationLengthAdjuster | None = None,
        system_instruction: Union[str, Iterable[TextBlockParam]] | NotGiven = NOT_GIVEN,
        temperature: float | NotGiven = NOT_GIVEN,
        top_k: int | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven = NOT_GIVEN,
    ):
        self.model_name = model_name
        self.api_type = api_type
        self.api_key_env_name = api_key_env_name
        self.auth_token_env_name = auth_token_env_name
        self.endpoint_env_name = endpoint_env_name
        self.aws_secret_key_env_name = aws_secret_key_env_name
        self.aws_access_key_env_name = aws_access_key_env_name
        self.aws_region_env_name = aws_region_env_name
        self.aws_session_token_env_name = aws_session_token_env_name
        self.timeout = timeout
        self.max_tokens = max_tokens
        self.max_retries = max_retries
        self.default_headers = default_headers
        self.default_query = default_query
        self.http_client = http_client
        self.transport = transport
        self.proxies = proxies
        self.connection_pool_limits = connection_pool_limits
        self._strict_response_validation = _strict_response_validation
        self.conversation_length_adjuster = conversation_length_adjuster
        self.system_instruction = system_instruction
        self.gc_region_env_name = gc_region_env_name
        self.gc_project_id_env_name = gc_project_id_env_name
        self.gc_access_token_env_name = gc_access_token_env_name
        self.credentials = credentials
        self.temperature = temperature
        self.top_k = top_k
        self.top_p = top_p

        ClientToolConfig = self._get_client_tool_type()
        client_tool_configs = ClientToolConfig.from_universal_configs(tool_configs)

        self.tools = self._set_runnable_tools_dict(tools)
        self.tool_configs = [f.format() for f in client_tool_configs]

    def _get_client_tool_type(self) -> ClaudeToolConfig:
        return ClaudeToolConfig

    def _get_tool_choice(self, tool_choice: str | None) -> ToolChoice:
        if tool_choice is None:
            return NOT_GIVEN
        elif tool_choice == "auto":
            return ToolChoiceToolChoiceAuto(type="auto")
        elif tool_choice == "any":
            return ToolChoiceToolChoiceAny(type="any")
        else:
            return ToolChoiceToolChoiceTool(type="tool", name=tool_choice)

    def run(
        self, messages: list[dict[str, Any]], tool_choice: str | None = None
    ) -> FunctionCallingResults:
        client = get_client(
            api_type=self.api_type,
            api_key_env_name=self.api_key_env_name,
            auth_token_env_name=self.auth_token_env_name,
            endpoint_env_name=self.endpoint_env_name,
            aws_secret_key_env_name=self.aws_secret_key_env_name,
            aws_access_key_env_name=self.aws_access_key_env_name,
            aws_region_env_name=self.aws_region_env_name,
            aws_session_token_env_name=self.aws_session_token_env_name,
            gc_region_env_name=self.gc_region_env_name,
            gc_project_id_env_name=self.gc_project_id_env_name,
            gc_access_token_env_name=self.gc_access_token_env_name,
            credentials=self.credentials,
            timeout=self.timeout,
            max_retries=self.max_retries,
            default_headers=self.default_headers,
            default_query=self.default_query,
            http_client=self.http_client,
            transport=self.transport,
            proxies=self.proxies,
            connection_pool_limits=self.connection_pool_limits,
            _strict_response_validation=self._strict_response_validation,
        )

        response = completion(
            client=client,
            model_name=self.model_name,
            messages=messages,
            system_instruction=self.system_instruction,
            max_tokens=self.max_tokens,
            timeout=self.timeout,
            tools=self.tool_configs,
            tool_choice=self._get_tool_choice(tool_choice),
            temperature=self.temperature,
            top_k=self.top_k,
            top_p=self.top_p,
        )

        contents = []
        calls = []
        model = response.model
        usage = Usage(
            model_name=model,
            prompt_tokens=response.usage.input_tokens,
            completion_tokens=response.usage.output_tokens,
        )

        for content in response.content:
            if isinstance(content, TextBlock):
                calls.append(
                    TextContent(
                        text=content.text,
                    )
                )
            elif isinstance(content, ToolUseBlock):
                args = content.input
                funcname = content.name
                call_id = content.id
                output = ToolOutput(
                    call_id=call_id,
                    funcname=funcname,
                    args=json.dumps(args),
                    output=str(self.tools[funcname](**args)),
                )

                contents.append(output)
                calls.append(
                    ToolCallResponse(
                        call_id=call_id,
                        name=funcname,
                        args=args,
                    )
                )

        return FunctionCallingResults(
            usage=usage,
            results=contents,
            prompt=copy.deepcopy(messages),
            calls=calls,
        )

    async def arun(
        self, messages: list[dict[str, Any]], tool_choice: str | None = None
    ) -> FunctionCallingResults:
        client = get_async_client(
            api_type=self.api_type,
            api_key_env_name=self.api_key_env_name,
            auth_token_env_name=self.auth_token_env_name,
            endpoint_env_name=self.endpoint_env_name,
            aws_secret_key_env_name=self.aws_secret_key_env_name,
            aws_access_key_env_name=self.aws_access_key_env_name,
            aws_region_env_name=self.aws_region_env_name,
            aws_session_token_env_name=self.aws_session_token_env_name,
            gc_region_env_name=self.gc_region_env_name,
            gc_project_id_env_name=self.gc_project_id_env_name,
            gc_access_token_env_name=self.gc_access_token_env_name,
            credentials=self.credentials,
            timeout=self.timeout,
            max_retries=self.max_retries,
            default_headers=self.default_headers,
            default_query=self.default_query,
            http_client=self.http_client,
            transport=self.transport,
            proxies=self.proxies,
            connection_pool_limits=self.connection_pool_limits,
            _strict_response_validation=self._strict_response_validation,
        )

        response = await acompletion(
            client=client,
            model_name=self.model_name,
            messages=messages,
            system_instruction=self.system_instruction,
            max_tokens=self.max_tokens,
            timeout=self.timeout,
            tools=self.tool_configs,
            tool_choice=self._get_tool_choice(tool_choice),
            temperature=self.temperature,
            top_k=self.top_k,
            top_p=self.top_p,
        )

        contents = []
        calls = []
        model = response.model
        usage = Usage(
            model_name=model,
            prompt_tokens=response.usage.input_tokens,
            completion_tokens=response.usage.output_tokens,
        )

        for content in response.content:
            if isinstance(content, TextBlock):
                calls.append(
                    TextContent(
                        text=content.text,
                    )
                )
            elif isinstance(content, ToolUseBlock):
                args = content.input
                funcname = content.name
                call_id = content.id
                output = ToolOutput(
                    call_id=call_id,
                    funcname=funcname,
                    args=json.dumps(args),
                    output=str(self.tools[funcname](**args)),
                )

                contents.append(output)
                calls.append(
                    ToolCallResponse(
                        call_id=call_id,
                        name=funcname,
                        args=args,
                    )
                )

        return FunctionCallingResults(
            usage=usage,
            results=contents,
            prompt=copy.deepcopy(messages),
            calls=calls,
        )


class AnthropicFunctionCallingModule(FunctionCallingWrapperModule):
    def __init__(
        self,
        model_name: str,
        tools: list[Callable],
        tool_configs: list[ToolConfig],
        api_type: str = "anthropic",
        api_key_env_name: str | None = None,
        auth_token_env_name: str | None = None,
        endpoint_env_name: str | httpx.URL | None = None,
        aws_secret_key_env_name: str | None = None,
        aws_access_key_env_name: str | None = None,
        aws_region_env_name: str | None = None,
        aws_session_token_env_name: str | None = None,
        gc_region_env_name: str | NotGiven = NOT_GIVEN,
        gc_project_id_env_name: str | NotGiven = NOT_GIVEN,
        gc_access_token_env_name: str | None = None,
        credentials: Any | None = None,
        timeout: Union[float, Timeout, None, NotGiven] = 600,
        max_retries: int = DEFAULT_MAX_RETRIES,
        default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        http_client: httpx.Client | None = None,
        transport: Transport | None = None,
        proxies: ProxiesTypes | None = None,
        connection_pool_limits: httpx.Limits | None = None,
        _strict_response_validation: bool = False,
        conversation_length_adjuster: BaseConversationLengthAdjuster | None = None,
        system_instruction: Union[str, Iterable[TextBlockParam]] | NotGiven = NOT_GIVEN,
        conversation_memory: BaseConversationMemory | None = None,
        content_filter: BaseFilter | None = None,
        token_counter: TokenCounter | None = None,
        temperature: float | NotGiven = NOT_GIVEN,
        top_k: int | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven = NOT_GIVEN,
    ):
        # The module to call client API
        function_calling_model = AnthropicFunctionCallingCoreModule(
            model_name=model_name,
            tools=tools,
            tool_configs=tool_configs,
            api_type=api_type,
            api_key_env_name=api_key_env_name,
            auth_token_env_name=auth_token_env_name,
            endpoint_env_name=endpoint_env_name,
            aws_secret_key_env_name=aws_secret_key_env_name,
            aws_access_key_env_name=aws_access_key_env_name,
            aws_region_env_name=aws_region_env_name,
            aws_session_token_env_name=aws_session_token_env_name,
            gc_region_env_name=gc_region_env_name,
            gc_project_id_env_name=gc_project_id_env_name,
            gc_access_token_env_name=gc_access_token_env_name,
            credentials=credentials,
            timeout=timeout,
            max_retries=max_retries,
            default_headers=default_headers,
            default_query=default_query,
            http_client=http_client,
            transport=transport,
            proxies=proxies,
            connection_pool_limits=connection_pool_limits,
            _strict_response_validation=_strict_response_validation,
            conversation_length_adjuster=conversation_length_adjuster,
            system_instruction=system_instruction,
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
        )

        super().__init__(
            function_calling_model=function_calling_model,
            conversation_memory=conversation_memory,
            content_filter=content_filter,
            token_counter=token_counter,
        )

    def _get_client_message_type(self) -> ClaudeMessage:
        return ClaudeMessage



import base64
import io
import random
import string
from functools import partial
from typing import Any, Generator, Iterable
from urllib.parse import urlparse

import numpy as np
from PIL import Image
from pydantic import BaseModel


def make_batch(
    iterable: Iterable[Any], batch_size: int = 1, overlap: int = 0
) -> Generator[Iterable[Any], None, None]:
    if overlap >= batch_size:
        raise ValueError("overlap must be less than batch_size")

    length = len(iterable)

    st: int = 0
    while st < length:
        en = min(st + batch_size, length)
        batch = iterable[st:en]
        yield batch

        if en == length:
            break

        st += batch_size - overlap


def pil2bytes(image: Image.Image) -> bytes:
    num_byteio = io.BytesIO()
    image.save(num_byteio, format=image.format.lower() if image.format else "jpeg")
    num_byteio.seek(0)
    image_bytes = num_byteio.getvalue()
    return image_bytes


def image2bytes(image: Image.Image | np.ndarray | bytes) -> bytes:
    if isinstance(image, Image.Image):
        return pil2bytes(image)
    elif isinstance(image, np.ndarray):
        image_pil = Image.fromarray(image)
        return pil2bytes(image_pil)
    elif isinstance(image, bytes):
        return image
    else:
        raise ValueError(f"Type of {type(image)} is not supported for image.")


def base64_encode(data: Any):
    return base64.b64encode(data)


def encode_image(image: Image.Image | np.ndarray | bytes, as_utf8: bool = False) -> str:
    image_bytes = image2bytes(image)
    encoded = base64_encode(image_bytes)

    if as_utf8:
        return encoded.decode("utf-8")
    else:
        return encoded


def decode_image(image_encoded: str, as_utf8: bool = False) -> Image.Image:
    if not as_utf8:
        image_encoded_utf = image_encoded
    else:
        image_encoded_utf = image_encoded.encode("utf-8")

    image_bytes = base64.b64decode(image_encoded_utf)
    byteio = io.BytesIO(image_bytes)
    return Image.open(byteio)


def generate_dummy_call_id(n: int) -> str:
    return "".join(random.choices(string.ascii_letters + string.digits, k=n))


def model2func(model: BaseModel):
    def run_model(model: BaseModel, **kwargs):
        return model(**kwargs).run()

    return partial(run_model, model=model)


def is_valid_uri(text: str) -> bool:
    """Function to determine if a given string is a URI

    Args:
        text (str): The string to check

    Returns:
        bool: True if the string is a URI, False otherwise
    """
    try:
        result = urlparse(text)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False



import os
from typing import Mapping

import httpx
from anthropic import AnthropicVertex, AsyncAnthropicVertex
from anthropic._base_client import DEFAULT_MAX_RETRIES
from anthropic._types import NOT_GIVEN, NotGiven, ProxiesTypes, Transport
from google.auth.credentials import Credentials as GoogleCredentials


def get_vertexai_client(
    gc_region_env_name: str | NotGiven = NOT_GIVEN,
    gc_project_id_env_name: str | NotGiven = NOT_GIVEN,
    gc_access_token_env_name: str | None = None,
    credentials: GoogleCredentials | None = None,
    endpoint_env_name: str | httpx.URL | None = None,
    timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    max_retries: int = DEFAULT_MAX_RETRIES,
    default_headers: Mapping[str, str] | None = None,
    default_query: Mapping[str, object] | None = None,
    http_client: httpx.Client | None = None,
    transport: Transport | None = None,
    proxies: ProxiesTypes | None = None,
    connection_pool_limits: httpx.Limits | None = None,
    _strict_response_validation: bool = False,
) -> AnthropicVertex:
    return AnthropicVertex(
        region=os.getenv(gc_region_env_name) if gc_region_env_name else None,
        project_id=os.getenv(gc_project_id_env_name) if gc_project_id_env_name else None,
        access_token=os.getenv(gc_access_token_env_name) if gc_access_token_env_name else None,
        credentials=credentials,
        base_url=os.getenv(endpoint_env_name) if endpoint_env_name else None,
        timeout=timeout,
        max_retries=max_retries,
        default_headers=default_headers,
        default_query=default_query,
        http_client=http_client,
        transport=transport,
        proxies=proxies,
        connection_pool_limits=connection_pool_limits,
        _strict_response_validation=_strict_response_validation,
    )


def get_async_vertexai_client(
    gc_region_env_name: str | NotGiven = NOT_GIVEN,
    gc_project_id_env_name: str | NotGiven = NOT_GIVEN,
    gc_access_token_env_name: str | None = None,
    credentials: GoogleCredentials | None = None,
    endpoint_env_name: str | httpx.URL | None = None,
    timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    max_retries: int = DEFAULT_MAX_RETRIES,
    default_headers: Mapping[str, str] | None = None,
    default_query: Mapping[str, object] | None = None,
    http_client: httpx.Client | None = None,
    transport: Transport | None = None,
    proxies: ProxiesTypes | None = None,
    connection_pool_limits: httpx.Limits | None = None,
    _strict_response_validation: bool = False,
) -> AsyncAnthropicVertex:
    return AsyncAnthropicVertex(
        region=os.getenv(gc_region_env_name) if gc_region_env_name else None,
        project_id=os.getenv(gc_project_id_env_name) if gc_project_id_env_name else None,
        access_token=os.getenv(gc_access_token_env_name) if gc_access_token_env_name else None,
        credentials=credentials,
        base_url=os.getenv(endpoint_env_name) if endpoint_env_name else None,
        timeout=timeout,
        max_retries=max_retries,
        default_headers=default_headers,
        default_query=default_query,
        http_client=http_client,
        transport=transport,
        proxies=proxies,
        connection_pool_limits=connection_pool_limits,
        _strict_response_validation=_strict_response_validation,
    )



import re
from typing import Any

from ...base import BaseMetadataFilter


class SQLiteMetadataFilter(BaseMetadataFilter):
    """
    Apply a WHERE clause in SQLite format to a metadata dictionary.
    """

    def __init__(self, where: str):
        where = re.sub("%", "", where)
        self.conditions: list[tuple[Any, str, Any]] = self._parse_where_clause(where_clause=where)

    def _parse_value(self, value):
        value = value.strip()
        if value.startswith("'") and value.endswith("'"):
            value = value[1:-1]  # Remove the single quotes
        elif value.startswith('"') and value.endswith('"'):
            value = value[1:-1]  # Remove the double quotes
        return value

    def _parse_in_clause(self, field, values, expression):
        values = values.strip()[1:-1]  # Remove the surrounding parentheses
        value_list = [self._parse_value(v.strip()) for v in values.split(",")]
        return [field, expression, value_list]

    def _replace_in_expression_and_append_clause(
        self, expression: str, where_clause: str, stack: list
    ):
        clauses = re.findall(
            r"\w+\s+{expression}\s+\([^)]*\)".format(expression=expression),
            where_clause,
            flags=re.IGNORECASE,
        )

        expression_ = expression.replace(" ", "_")
        for clause in clauses:
            where_clause = where_clause.replace(clause, f"{expression_}_CLAUSE_{len(stack)}")
            stack.append(clause)

        return where_clause, stack

    def _replace_is_not_expression(self, expression: str, where_clause: str):
        where_clause = where_clause.replace(expression, expression.replace(" ", "_"))
        return where_clause

    def _parse_where_clause(self, where_clause: str) -> list[tuple[Any, str, Any]]:
        conditions = []
        stack = []

        # Separate typical clause to distinguish brackets in IN clause from other brackets
        where_clause, stack = self._replace_in_expression_and_append_clause(
            "NOT IN", where_clause, stack
        )
        where_clause, stack = self._replace_in_expression_and_append_clause(
            "IN", where_clause, stack
        )

        # Replace IS NOT expressions with IS_NOT
        where_clause = self._replace_is_not_expression("IS NOT", where_clause)

        tokens = re.split(r"(\(|\)|\s+AND\s+|\s+OR\s+)", where_clause, flags=re.IGNORECASE)
        tokens = [token.strip() for token in tokens if token.strip()]

        for token in tokens:
            if token == "(":
                stack.append(conditions)
                conditions = []
            elif token == ")":
                last_conditions = conditions
                conditions = stack.pop()
                conditions.append(last_conditions)
            elif token.upper() in ["AND", "OR"]:
                conditions.append(token.upper())
            elif token.startswith("IN_CLAUSE_"):
                index = int(token.split("_")[-1])
                in_clause = stack[index]
                in_match = re.match(r"(\w+)\s+IN\s+(\(.+\))", in_clause, re.IGNORECASE)
                if in_match:
                    field, values = in_match.groups()
                    conditions.append(self._parse_in_clause(field, values, "IN"))
            elif token.startswith("NOT_IN_CLAUSE_"):
                index = int(token.split("_")[-1])
                not_in_clause = stack[index]
                not_in_match = re.match(r"(\w+)\s+NOT IN\s+(\(.+\))", not_in_clause, re.IGNORECASE)
                if not_in_match:
                    field, values = not_in_match.groups()
                    conditions.append(self._parse_in_clause(field, values, "NOT IN"))
            elif "IS_NOT" in token:
                field, value = token.split(" IS_NOT ")
                value = self._parse_value(value)
                conditions.append([field, "IS NOT", value])
            else:
                match = re.match(
                    r"(\w+)\s*(>=|<=|!=|<>|=|==|>|<|LIKE|IS)\s*(.+)", token, re.IGNORECASE
                )
                if match:
                    field, operator, value = match.groups()
                    value = self._parse_value(value)
                    conditions.append([field, operator.upper(), value])

        return conditions

    def _apply_conditions(
        self, item: dict[str, Any], conditions: list[tuple[Any, str, Any]]
    ) -> bool:
        stack = []
        for condition in conditions:
            if condition in ["AND", "OR"]:
                stack.append(condition)
            else:
                if isinstance(condition, list) and ("AND" in condition or "OR" in condition):
                    stack.append(self._apply_conditions(item, condition))
                else:
                    field: str
                    operator: str
                    value: str
                    field, operator, value = condition
                    item_value = item.get(field)

                    if item_value is None:
                        result = False
                    else:
                        if operator == "=" or operator == "==" or operator == "IS":
                            if value.isdigit():
                                result = float(item_value) == float(value)
                            else:
                                result = item_value == value
                        elif operator == "!=" or operator == "<>" or operator == "IS NOT":
                            if value.isdigit():
                                result = float(item_value) != float(value)
                            else:
                                result = item_value != value
                        elif operator == ">":
                            if value.isdigit():
                                result = float(item_value) > float(value)
                            else:
                                result = item_value > value
                        elif operator == "<":
                            if value.isdigit():
                                result = float(item_value) < float(value)
                            else:
                                result = item_value < value
                        elif operator == ">=":
                            if value.isdigit():
                                result = float(item_value) >= float(value)
                            else:
                                result = item_value >= value
                        elif operator == "<=":
                            if value.isdigit():
                                result = float(item_value) <= float(value)
                            else:
                                result = item_value <= value
                        elif operator == "LIKE":
                            pattern_split = re.compile(r",|、|\s|　")
                            values = pattern_split.split(value)
                            result = any([bool(re.search(v, item_value)) for v in values if v])
                        elif operator == "NOT IN":
                            pattern_split = re.compile(r",|、|\s|　")
                            item_values = pattern_split.split(item_value)
                            value = set(value)
                            result = all([v not in value for v in item_values])
                        elif operator == "IN":
                            pattern_split = re.compile(r",|、|\s|　")
                            item_values = pattern_split.split(item_value)
                            value = set(value)
                            result = any([v in value for v in item_values])

                    stack.append(result)

        # Evaluate the condition stack
        while len(stack) > 1:
            left = stack.pop(0)
            operator = stack.pop(0)
            right = stack.pop(0)
            if operator == "AND":
                stack.insert(0, left and right)
            elif operator == "OR":
                stack.insert(0, left or right)

        if stack:
            return stack[0]
        else:
            return False

    def run(self, metadata: dict[str, Any]) -> bool:
        return self._apply_conditions(metadata, self.conditions)



from pathlib import Path

from PIL import Image

BASE_DIR = Path(__file__).parent


class SampleData:
    @staticmethod
    def load_image() -> Image.Image:
        return Image.open(BASE_DIR.parent.parent / "data" / "sample.jpg")



import json
import os

from ..base import BaseConversationMemory


class JSONConversationMemory(BaseConversationMemory):
    def __init__(self, path: str = "conversation_memory.json", exist_ok: bool = False):
        self.path = path

        if not exist_ok and os.path.exists(self.path):
            os.remove(self.path)

    def store(self, conversation_history: list[dict[str, str]]):
        with open(self.path, "w") as f:
            json.dump(conversation_history, f, ensure_ascii=False)

    def load(self) -> list[dict[str, str]]:
        if os.path.exists(self.path):
            with open(self.path, "r") as f:
                conversation_history = json.load(f)
            return conversation_history
        else:
            return []
